#!/usr/bin/env python3
import json
import numpy as np
from pathlib import Path
from .flash_entropy_search_core import FlashEntropySearchCore


class FlashEntropySearchCoreLowMemory(FlashEntropySearchCore):
    def __init__(self, path_data, max_ms2_tolerance_in_da=0.024, mz_index_step=0.0001, intensity_weight="entropy") -> None:
        """
        Initialize the EntropySearch class.
        This class use file.read function to read the data from the file, which is suitable for very low memory usage.
        
        :param path_data:   The path to save the index data.
        :param max_ms2_tolerance_in_da:  The maximum MS2 tolerance in Da.
        :param mz_index_step:   The step size for the m/z index.
        :param intensity_weight:    The weight for the intensity in the entropy calculation, can be "entropy" or None. Default is "entropy".
            - None: The intensity will not be weighted, then the unweighted similarity will be calculated.
            - "entropy": The intensity will be weighted by the entropy, then the entropy similarity will be calculated.
        """
        super().__init__(max_ms2_tolerance_in_da=max_ms2_tolerance_in_da, mz_index_step=mz_index_step, intensity_weight=intensity_weight)
        self.path_data = Path(str(path_data))
        self.path_data.mkdir(parents=True, exist_ok=True)
        self.index_file = []

    def __del__(self):
        for file in self.index_file:
            file.close()

    def _generate_index_from_peak_data(self, peak_data, max_indexed_mz, append):
        total_peaks_num = peak_data.shape[0]

        # Sort with precursor m/z.
        peak_data.sort(order="ion_mz")

        # Record the m/z, intensity, and spectrum index information for product ions.
        (peak_data["ion_mz"]).tofile(self.path_data / "all_ions_mz.npy")
        (peak_data["intensity"]).tofile(self.path_data / "all_ions_intensity.npy")
        (peak_data["spec_idx"]).tofile(self.path_data / "all_ions_spec_idx.npy")

        # all_ions_mz = self._convert_view_to_array(peak_data.view(np.float32).reshape(total_peaks_num, -1)[:, 0], np.float32, "all_ions_mz")
        # all_ions_intensity = self._convert_view_to_array(peak_data.view(np.float32).reshape(total_peaks_num, -1)[:, 2], np.float32, "all_ions_intensity")
        # all_ions_spec_idx = self._convert_view_to_array(peak_data.view(np.uint32).reshape(total_peaks_num, -1)[:, 3], np.uint32, "all_ions_spec_idx")

        # Assign the index of the product ions.
        peak_data["peak_idx"] = np.arange(0, self.total_peaks_num, dtype=np.uint64)

        # Build index for fast access to the ion's m/z.
        all_ions_mz = np.memmap(self.path_data / "all_ions_mz.npy", dtype=np.float32, mode="r", shape=(total_peaks_num,))
        max_mz = min(np.max(all_ions_mz), max_indexed_mz)
        search_array = np.arange(0.0, max_mz, self.mz_index_step)
        all_ions_mz_idx_start = np.searchsorted(all_ions_mz, search_array, side="left").astype(np.int64)
        all_ions_mz_idx_start.tofile(self.path_data / "all_ions_mz_idx_start.npy")

        ############## Step 3: Build the index by sort with neutral loss mass. ##############
        # Sort with the neutral loss mass.
        peak_data.sort(order="nl_mass")

        # Record the m/z, intensity, spectrum index, and product ions index information for neutral loss ions.
        (peak_data["nl_mass"]).tofile(self.path_data / "all_nl_mass.npy")
        (peak_data["intensity"]).tofile(self.path_data / "all_nl_intensity.npy")
        (peak_data["spec_idx"]).tofile(self.path_data / "all_nl_spec_idx.npy")
        (peak_data["peak_idx"]).tofile(self.path_data / "all_ions_idx_for_nl.npy")

        # all_nl_mass = self._convert_view_to_array(peak_data.view(np.float32).reshape(total_peaks_num, -1)[:, 1], np.float32, "all_nl_mass")
        # all_nl_intensity = self._convert_view_to_array(peak_data.view(np.float32).reshape(total_peaks_num, -1)[:, 2], np.float32, "all_nl_intensity")
        # all_nl_spec_idx = self._convert_view_to_array(peak_data.view(np.uint32).reshape(total_peaks_num, -1)[:, 3], np.uint32, "all_nl_spec_idx")
        # all_ions_idx_for_nl = self._convert_view_to_array(peak_data.view(np.uint64).reshape(total_peaks_num, -1)[:, 2], np.uint64, "all_ions_idx_for_nl")

        # Build the index for fast access to the neutral loss mass.
        all_nl_mass = np.memmap(self.path_data / "all_nl_mass.npy", dtype=np.float32, mode="r", shape=(total_peaks_num,))
        max_mz = min(np.max(all_nl_mass), max_indexed_mz)
        search_array = np.arange(0.0, max_mz, self.mz_index_step)
        all_nl_mass_idx_start = np.searchsorted(all_nl_mass, search_array, side="left").astype(np.int64)
        all_nl_mass_idx_start.tofile(self.path_data / "all_nl_mass_idx_start.npy")

        ############## Step 4: Save the index. ##############
        self.write()
        self.read()
        return self.index

    def read(self, path_data=None):
        """
        Read the index from the file.
        """
        if path_data is not None:
            self.path_data = Path(path_data)

        try:
            self.index = []
            for name in self.index_names:
                self.index.append(np.memmap(self.path_data / f"{name}.npy", dtype=self.index_dtypes[name], mode="r"))
                file_cur = open(self.path_data / f"{name}.npy", "rb")
                file_cur.data_type = np.dtype(self.index_dtypes[name])
                self.index_file.append(file_cur)

            with open(self.path_data / "information.json", "r") as f:
                information = json.load(f)
            self.mz_index_step = information["mz_index_step"]
            self.total_spectra_num = information["total_spectra_num"]
            self.total_peaks_num = information["total_peaks_num"]
            self.max_ms2_tolerance_in_da = information["max_ms2_tolerance_in_da"]
            return True
        except:
            return False

    def write(self, path_data=None):
        """
        Write the index to the file.
        """
        if path_data is not None:
            assert Path(path_data) == self.path_data, "The path_data is not the same as the path_data in the class."

        information = {
            "mz_index_step": float(self.mz_index_step),
            "total_spectra_num": int(self.total_spectra_num),
            "total_peaks_num": int(self.total_peaks_num),
            "max_ms2_tolerance_in_da": float(self.max_ms2_tolerance_in_da),
        }
        json.dump(information, open(self.path_data / "information.json", "w"))

    def search_hybrid(self, target="cpu", precursor_mz=None, peaks=None, ms2_tolerance_in_da=0.02):
        """
        Perform the hybrid search for the MS/MS spectra.

        :param target: The target to perform the search. "cpu" for CPU, "gpu" for GPU.
        :param precursor_mz: The precursor m/z of the MS/MS spectra.
        :param peaks: The peaks of the MS/MS spectra, needs to be cleaned with the "clean_spectrum" function.
        :param ms2_tolerance_in_da: The MS/MS tolerance in Da.
        """
        if not self.index:
            return np.zeros(0, dtype=np.float32)
        if len(peaks) == 0:
            return np.zeros(self.total_spectra_num, dtype=np.float32)

        # Check peaks
        assert ms2_tolerance_in_da <= self.max_ms2_tolerance_in_da, "The MS2 tolerance is larger than the maximum MS2 tolerance."
        assert abs(np.sum(peaks[:, 1]) - 1) < 1e-4, "The peaks are not normalized to sum to 1."
        assert (
            peaks.shape[0] <= 1 or np.min(peaks[1:, 0] - peaks[:-1, 0]) > self.max_ms2_tolerance_in_da * 2
        ), "The peaks array should be sorted by m/z, and the m/z difference between two adjacent peaks should be larger than 2 * max_ms2_tolerance_in_da."
        (
            all_ions_mz_idx_start,
            all_ions_mz,
            all_ions_intensity,
            all_ions_spec_idx,
            all_nl_mass_idx_start,
            all_nl_mass,
            all_nl_intensity,
            all_nl_spec_idx,
            all_ions_idx_for_nl,
        ) = self.index
        (
            file_all_ions_mz_idx_start,
            file_all_ions_mz,
            file_all_ions_intensity,
            file_all_ions_spec_idx,
            file_all_nl_mass_idx_start,
            file_all_nl_mass,
            file_all_nl_intensity,
            file_all_nl_spec_idx,
            file_all_ions_idx_for_nl,
        ) = self.index_file
        index_number_in_one_da = int(1 / self.mz_index_step)

        # Prepare the query spectrum
        peaks = self._preprocess_peaks(peaks)

        # Go through all peak in the spectrum and determine the mz index range
        product_peak_match_idx_min = np.zeros(peaks.shape[0], dtype=np.uint64)
        product_peak_match_idx_max = np.zeros(peaks.shape[0], dtype=np.uint64)
        for peak_idx, (mz_query, _) in enumerate(peaks):
            # Determine the mz index range
            product_mz_idx_min = self._find_location_from_array_with_index(
                mz_query - ms2_tolerance_in_da, all_ions_mz, all_ions_mz_idx_start, "left", index_number_in_one_da
            )
            product_mz_idx_max = self._find_location_from_array_with_index(
                mz_query + ms2_tolerance_in_da, all_ions_mz, all_ions_mz_idx_start, "right", index_number_in_one_da
            )

            product_peak_match_idx_min[peak_idx] = product_mz_idx_min
            product_peak_match_idx_max[peak_idx] = product_mz_idx_max

        if target == "cpu":
            entropy_similarity = np.zeros(self.total_spectra_num, dtype=np.float32)
            # Go through all the peaks in the spectrum and calculate the entropy similarity
            for peak_idx, (mz, intensity) in enumerate(peaks):
                ###############################################################
                # Match the original product ion
                product_mz_idx_min = product_peak_match_idx_min[peak_idx]
                product_mz_idx_max = product_peak_match_idx_max[peak_idx]

                # Calculate the entropy similarity for this matched peak
                # modified_idx_product = all_ions_spec_idx[product_mz_idx_min:product_mz_idx_max]
                modified_idx_product = _read_data_from_file(file_all_ions_spec_idx, product_mz_idx_min, product_mz_idx_max)

                # modified_value_product = all_ions_intensity[product_mz_idx_min:product_mz_idx_max]
                modified_value_product = _read_data_from_file(file_all_ions_intensity, product_mz_idx_min, product_mz_idx_max)
                modified_value_product = self._score_peaks_with_cpu(intensity, modified_value_product)

                entropy_similarity[modified_idx_product] += modified_value_product

                ###############################################################
                # Match the neutral loss ions
                mz_nl = precursor_mz - mz
                # Determine the mz index range
                neutral_loss_mz_idx_min = self._find_location_from_array_with_index(
                    mz_nl - ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "left", index_number_in_one_da
                )
                neutral_loss_mz_idx_max = self._find_location_from_array_with_index(
                    mz_nl + ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "right", index_number_in_one_da
                )

                # Calculate the entropy similarity for this matched peak
                # modified_idx_nl = all_nl_spec_idx[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]
                modified_idx_nl = _read_data_from_file(file_all_nl_spec_idx, neutral_loss_mz_idx_min, neutral_loss_mz_idx_max)
                # modified_value_nl = all_nl_intensity[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]
                modified_value_nl = _read_data_from_file(file_all_nl_intensity, neutral_loss_mz_idx_min, neutral_loss_mz_idx_max)
                modified_value_nl = self._score_peaks_with_cpu(intensity, modified_value_nl)

                # Check if the neutral loss ion is already matched to other query peak as a product ion
                nl_matched_product_ion_idx = _read_data_from_file(file_all_ions_idx_for_nl, neutral_loss_mz_idx_min, neutral_loss_mz_idx_max)
                s1 = np.searchsorted(product_peak_match_idx_min, nl_matched_product_ion_idx, side="right")
                s2 = np.searchsorted(product_peak_match_idx_max - 1, nl_matched_product_ion_idx, side="left")
                modified_value_nl[s1 > s2] = 0

                # Check if this query peak is already matched to a product ion in the same library spectrum
                duplicate_idx_in_nl = self._remove_duplicate_with_cpu(modified_idx_product, modified_idx_nl, self.total_spectra_num)
                modified_value_nl[duplicate_idx_in_nl] = 0

                entropy_similarity[modified_idx_nl] += modified_value_nl
            return entropy_similarity

        elif target == "gpu":
            import cupy as cp

            entropy_transform = cp.ElementwiseKernel(
                "T intensity_a, T intensity_b",
                "T similarity",
                """T intensity_ab = intensity_a + intensity_b;
                similarity = intensity_ab * log2f(intensity_ab) - intensity_a * log2f(intensity_a) - intensity_b * log2f(intensity_b);""",
            )
            product_peak_match_idx_min_gpu = cp.array(product_peak_match_idx_min)
            product_peak_match_idx_max_gpu = cp.array(product_peak_match_idx_max - 1)

            entropy_similarity_modification_list = []
            # Go through all the peaks in the spectrum and calculate the entropy similarity
            for peak_idx, (mz, intensity) in enumerate(peaks):
                ###############################################################
                # Match the original product ion
                product_mz_idx_min = product_peak_match_idx_min[peak_idx]
                product_mz_idx_max = product_peak_match_idx_max[peak_idx]

                # Calculate the entropy similarity for this matched peak
                # modified_idx_product = all_ions_spec_idx[product_mz_idx_min:product_mz_idx_max]
                modified_idx_product = _read_data_from_file(file_all_ions_spec_idx, product_mz_idx_min, product_mz_idx_max)

                # modified_value_product = all_ions_intensity[product_mz_idx_min:product_mz_idx_max]
                modified_value_product = _read_data_from_file(file_all_ions_intensity, product_mz_idx_min, product_mz_idx_max)
                modified_value_product = self._score_peaks_gpu(entropy_transform, intensity, cp.array(modified_value_product))

                entropy_similarity_modification_list.append((modified_idx_product, modified_value_product.get()))
                del modified_value_product

                ###############################################################
                # Match the neutral loss ions
                mz_nl = precursor_mz - mz
                # Determine the mz index range
                neutral_loss_mz_idx_min = self._find_location_from_array_with_index(
                    mz_nl - ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "left", index_number_in_one_da
                )
                neutral_loss_mz_idx_max = self._find_location_from_array_with_index(
                    mz_nl + ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "right", index_number_in_one_da
                )
                # print(product_mz_idx_max - product_mz_idx_min, neutral_loss_mz_idx_max - neutral_loss_mz_idx_min)

                # Calculate the entropy similarity for this matched peak
                # modified_idx_nl = all_nl_spec_idx[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]
                modified_idx_nl = _read_data_from_file(file_all_nl_spec_idx, neutral_loss_mz_idx_min, neutral_loss_mz_idx_max)
                # modified_value_nl = all_nl_intensity[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]
                modified_value_nl = _read_data_from_file(file_all_nl_intensity, neutral_loss_mz_idx_min, neutral_loss_mz_idx_max)
                modified_value_nl = self._score_peaks_gpu(entropy_transform, intensity, cp.array(modified_value_nl))

                # Check if the neutral loss ion is already matched to other query peak as a product ion
                # nl_matched_product_ion_idx = cp.array(all_ions_idx_for_nl[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max])
                nl_matched_product_ion_idx = cp.array(_read_data_from_file(file_all_ions_idx_for_nl, neutral_loss_mz_idx_min, neutral_loss_mz_idx_max))
                s1 = cp.searchsorted(product_peak_match_idx_min_gpu, nl_matched_product_ion_idx, side="right")
                s2 = cp.searchsorted(product_peak_match_idx_max_gpu, nl_matched_product_ion_idx, side="left")
                modified_value_nl[s1 > s2] = 0

                # Check if this query peak is already matched to a product ion in the same library spectrum
                duplicate_idx_in_nl = self._remove_duplicate_with_gpu(modified_idx_product, modified_idx_nl, self.total_spectra_num)
                modified_value_nl[duplicate_idx_in_nl] = 0

                entropy_similarity_modification_list.append((modified_idx_nl, modified_value_nl.get()))
                del modified_value_nl, nl_matched_product_ion_idx, duplicate_idx_in_nl

            # Release the GPU memory
            cp_dmp = cp.get_default_memory_pool()
            _ = (cp_dmp.used_bytes(), cp_dmp.total_bytes(), cp_dmp.free_bytes())
            cp_dmp.free_all_blocks()
            # print(cp_dmp.used_bytes(), cp_dmp.total_bytes(), cp_dmp.free_bytes(), cp_dmp.get_limit())
            cp.get_default_pinned_memory_pool().free_all_blocks()
            # print(cp_dmp.used_bytes(), cp_dmp.total_bytes(), cp_dmp.free_bytes(), cp_dmp.get_limit())
            entropy_similarity = cp.zeros(self.total_spectra_num, dtype=np.float16)
            for modified_idx, modified_value in entropy_similarity_modification_list:
                entropy_similarity.scatter_add(cp.array(modified_idx), cp.array(modified_value))
            return entropy_similarity.get()
        else:
            raise ValueError("target should be cpu or gpu")


def _read_data_from_file(file_data, item_start, item_end):
    array_type = file_data.data_type
    type_size = array_type.itemsize
    file_data.seek(int(item_start * type_size))
    data = file_data.read(int(type_size * (item_end - item_start)))
    array = np.frombuffer(data, dtype=array_type)
    return array



#!/usr/bin/env python3
import json
import numpy as np
from pathlib import Path
from functools import reduce
import multiprocessing
from ..spectra import apply_weight_to_intensity


class FlashEntropySearchCoreForDynamicIndexing:
    def __init__(
        self,
        path_data,
        max_ms2_tolerance_in_da=0.024,
        mz_index_step=0.0001,
        intensity_weight="entropy",  # "entropy" or None
    ) -> None:
        """
        Initialize the EntropySearch class.

        :param path_data: The path array of the index files.
        :param max_ms2_tolerance_in_da: The maximum MS2 tolerance used when searching the MS/MS spectra, in Dalton. Default is 0.024.
        :param mz_index_step:   The step size of the m/z index, in Dalton. Default is 0.0001.
                                The smaller the step size, the faster the search, but the larger the index size and longer the index building time.
        :param intensity_weight: The weight of the intensity, can be "entropy" or None. If set to "entropy", the intensity will be weighted by the entropy.
                                If set to None, the intensity will not be weighted, which is equivalent to the unweighted entropy similarity.
        """
        self.mz_index_step = mz_index_step
        self._init_for_multiprocessing = False
        self.max_ms2_tolerance_in_da = max_ms2_tolerance_in_da
        self.intensity_weight = intensity_weight

        self.total_spectra_num = 0
        self.total_peaks_num = 0
        self.index = []

        self.path_data = Path(path_data)

    def search(
        self,
        method="open",
        target="cpu",
        precursor_mz=None,
        peaks=None,
        ms2_tolerance_in_da=0.02,
        output_matched_peak_number=False,
    ):
        """
        Perform identity-, open- or neutral loss search on the MS/MS spectra library.

        :param method:  The search method, can be "open" or "neutral_loss".
                        Set it to "open" for identity search and open search, set it to "neutral_loss" for neutral loss search.
        :param target:  The target to search, can be "cpu" or "gpu".
        :param precursor_mz:    The precursor m/z of the query MS/MS spectrum, required for neutral loss search.
        :param peaks:   The peaks of the query MS/MS spectrum. The peaks need to be precleaned by "clean_spectrum" function.
        :param ms2_tolerance_in_da: The MS2 tolerance used when searching the MS/MS spectra, in Dalton. Default is 0.02.
        :param output_matched_peak_number: Whether to output the number of matched peaks. Only supported when target is "cpu".
                                            If set to True, the function will return a tuple of (entropy_similarity, matched_peak_number).
        """
        if not self.index:
            return np.zeros(0, dtype=np.float32)
        if len(peaks) == 0:
            return np.zeros(self.total_spectra_num, dtype=np.float32)

        # Check peaks
        assert ms2_tolerance_in_da <= self.max_ms2_tolerance_in_da, "The MS2 tolerance is larger than the maximum MS2 tolerance."
        assert abs(np.sum(peaks[:, 1]) - 1) < 1e-4, "The peaks are not normalized to sum to 1."
        assert (
            peaks.shape[0] <= 1 or np.min(peaks[1:, 0] - peaks[:-1, 0]) > self.max_ms2_tolerance_in_da * 2
        ), "The peaks array should be sorted by m/z, and the m/z difference between two adjacent peaks should be larger than 2 * max_ms2_tolerance_in_da."
        index_number_in_one_da = int(1 / self.mz_index_step)

        # Prepare the query spectrum
        peaks = self._preprocess_peaks(peaks)

        # Prepare the library
        if method == "open":
            library_mz_idx_start, library_mz, library_peaks_intensity, library_spec_idx = self.index[0]
        elif method == "neutral_loss":
            if self.index[1] is None:
                raise RuntimeError("The index for neutral loss search is not built.")
            library_mz_idx_start, library_mz, library_peaks_intensity, library_spec_idx, _ = self.index[1]
            peaks[:, 0] = precursor_mz - peaks[:, 0]

        # Start searching
        if target == "cpu":
            entropy_similarity = np.zeros(self.total_spectra_num, dtype=np.float32)
            if output_matched_peak_number:
                matched_peak_number = np.zeros(self.total_spectra_num, dtype=np.uint16)
        else:
            import cupy as cp

            entropy_transform = cp.ElementwiseKernel(
                "T intensity_a, T intensity_b",
                "T similarity",
                """T intensity_ab = intensity_a + intensity_b;
                similarity = intensity_ab * log2f(intensity_ab) - intensity_a * log2f(intensity_a) - intensity_b * log2f(intensity_b);""",
            )
            entropy_similarity = cp.zeros(self.total_spectra_num, dtype=np.float32)

        # Go through all the peaks in the spectrum
        for mz_query, intensity_query in peaks:
            # Determine the mz index range
            product_mz_idx_min = self._find_location_from_array_with_index(
                mz_query - ms2_tolerance_in_da, library_mz, library_mz_idx_start, "left", index_number_in_one_da
            )
            product_mz_idx_max = self._find_location_from_array_with_index(
                mz_query + ms2_tolerance_in_da, library_mz, library_mz_idx_start, "right", index_number_in_one_da
            )

            if target == "cpu":
                intensity_library = library_peaks_intensity[product_mz_idx_min:product_mz_idx_max]
                modified_idx = library_spec_idx[product_mz_idx_min:product_mz_idx_max]
                entropy_similarity[modified_idx] += self._score_peaks_with_cpu(intensity_query, intensity_library)
                if output_matched_peak_number:
                    matched_peak_number[modified_idx] += 1
            elif target == "gpu":
                intensity_library = cp.array(library_peaks_intensity[product_mz_idx_min:product_mz_idx_max])
                modified_value = entropy_transform(intensity_library, intensity_query)
                modified_idx = cp.array(library_spec_idx[product_mz_idx_min:product_mz_idx_max])
                entropy_similarity.scatter_add(modified_idx, modified_value)

        if target == "cpu":
            if output_matched_peak_number:
                return entropy_similarity, matched_peak_number
            else:
                return entropy_similarity
        elif target == "gpu":
            entropy_similarity = entropy_similarity.get()
            return entropy_similarity

    def search_hybrid(self, target="cpu", precursor_mz=None, peaks=None, ms2_tolerance_in_da=0.02):
        """
        Perform the hybrid search for the MS/MS spectra.

        :param target: The target to perform the search. "cpu" for CPU, "gpu" for GPU.
        :param precursor_mz: The precursor m/z of the MS/MS spectra.
        :param peaks: The peaks of the MS/MS spectra, needs to be cleaned with the "clean_spectrum" function.
        :param ms2_tolerance_in_da: The MS/MS tolerance in Da.
        """
        if not self.index:
            return np.zeros(0, dtype=np.float32)
        if len(peaks) == 0:
            return np.zeros(self.total_spectra_num, dtype=np.float32)

        # Check peaks
        assert ms2_tolerance_in_da <= self.max_ms2_tolerance_in_da, "The MS2 tolerance is larger than the maximum MS2 tolerance."
        assert abs(np.sum(peaks[:, 1]) - 1) < 1e-4, "The peaks are not normalized to sum to 1."
        assert (
            peaks.shape[0] <= 1 or np.min(peaks[1:, 0] - peaks[:-1, 0]) > self.max_ms2_tolerance_in_da * 2
        ), "The peaks array should be sorted by m/z, and the m/z difference between two adjacent peaks should be larger than 2 * max_ms2_tolerance_in_da."
        index_number_in_one_da = int(1 / self.mz_index_step)
        fragment_ion_index, neutral_loss_index = self.index
        all_ions_mz_idx_start, all_ions_mz, all_ions_intensity, all_ions_spec_idx = fragment_ion_index
        all_nl_mass_idx_start, all_nl_mass, all_nl_intensity, all_nl_spec_idx, ions_mz_for_nl = neutral_loss_index
        # Prepare the query spectrum
        peaks = self._preprocess_peaks(peaks)

        # Go through all peak in the spectrum and determine the mz index range
        product_peak_match_mz_min = peaks[:, 0] - ms2_tolerance_in_da
        product_peak_match_mz_max = peaks[:, 0] + ms2_tolerance_in_da

        product_peak_match_idx_min = np.zeros(peaks.shape[0], dtype=np.uint64)
        product_peak_match_idx_max = np.zeros(peaks.shape[0], dtype=np.uint64)
        for peak_idx, (mz_query, _) in enumerate(peaks):
            # Determine the mz index range
            product_mz_idx_min = self._find_location_from_array_with_index(
                mz_query - ms2_tolerance_in_da, all_ions_mz, all_ions_mz_idx_start, "left", index_number_in_one_da
            )
            product_mz_idx_max = self._find_location_from_array_with_index(
                mz_query + ms2_tolerance_in_da, all_ions_mz, all_ions_mz_idx_start, "right", index_number_in_one_da
            )

            product_peak_match_idx_min[peak_idx] = product_mz_idx_min
            product_peak_match_idx_max[peak_idx] = product_mz_idx_max

        if target == "cpu":
            entropy_similarity = np.zeros(self.total_spectra_num, dtype=np.float32)
            # Go through all the peaks in the spectrum and calculate the entropy similarity
            for peak_idx, (mz, intensity) in enumerate(peaks):
                ###############################################################
                # Match the original product ion
                product_mz_idx_min = product_peak_match_idx_min[peak_idx]
                product_mz_idx_max = product_peak_match_idx_max[peak_idx]

                # Calculate the entropy similarity for this matched peak
                modified_idx_product = all_ions_spec_idx[product_mz_idx_min:product_mz_idx_max]
                modified_value_product = self._score_peaks_with_cpu(intensity, all_ions_intensity[product_mz_idx_min:product_mz_idx_max])

                entropy_similarity[modified_idx_product] += modified_value_product

                ###############################################################
                # Match the neutral loss ions
                mz_nl = precursor_mz - mz
                # Determine the mz index range
                neutral_loss_mz_idx_min = self._find_location_from_array_with_index(
                    mz_nl - ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "left", index_number_in_one_da
                )
                neutral_loss_mz_idx_max = self._find_location_from_array_with_index(
                    mz_nl + ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "right", index_number_in_one_da
                )

                # Calculate the entropy similarity for this matched peak
                modified_idx_nl = all_nl_spec_idx[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]
                modified_value_nl = self._score_peaks_with_cpu(intensity, all_nl_intensity[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max])

                # Check if the neutral loss ion is already matched to other query peak as a product ion
                nl_matched_product_ion_mz = ions_mz_for_nl[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]
                s1 = np.searchsorted(product_peak_match_mz_min, nl_matched_product_ion_mz, side="right")
                s2 = np.searchsorted(product_peak_match_mz_max, nl_matched_product_ion_mz, side="left")
                modified_value_nl[s1 > s2] = 0

                # Check if this query peak is already matched to a product ion in the same library spectrum
                duplicate_idx_in_nl = self._remove_duplicate_with_cpu(modified_idx_product, modified_idx_nl, self.total_spectra_num)
                modified_value_nl[duplicate_idx_in_nl] = 0

                entropy_similarity[modified_idx_nl] += modified_value_nl
            return entropy_similarity

        elif target == "gpu":
            import cupy as cp

            entropy_transform = cp.ElementwiseKernel(
                "T intensity_a, T intensity_b",
                "T similarity",
                """T intensity_ab = intensity_a + intensity_b;
                similarity = intensity_ab * log2f(intensity_ab) - intensity_a * log2f(intensity_a) - intensity_b * log2f(intensity_b);""",
            )
            product_peak_match_mz_min_gpu = cp.array(product_peak_match_mz_min)
            product_peak_match_mz_max_gpu = cp.array(product_peak_match_mz_max)

            entropy_similarity_modification_list = []
            # Go through all the peaks in the spectrum and calculate the entropy similarity
            for peak_idx, (mz, intensity) in enumerate(peaks):
                ###############################################################
                # Match the original product ion
                product_mz_idx_min = product_peak_match_idx_min[peak_idx]
                product_mz_idx_max = product_peak_match_idx_max[peak_idx]

                # Calculate the entropy similarity for this matched peak
                modified_idx_product = all_ions_spec_idx[product_mz_idx_min:product_mz_idx_max]
                modified_value_product = self._score_peaks_gpu(
                    entropy_transform, intensity, cp.array(all_ions_intensity[product_mz_idx_min:product_mz_idx_max])
                )

                entropy_similarity_modification_list.append((modified_idx_product, modified_value_product.get()))
                del modified_value_product

                ###############################################################
                # Match the neutral loss ions
                mz_nl = precursor_mz - mz
                # Determine the mz index range
                neutral_loss_mz_idx_min = self._find_location_from_array_with_index(
                    mz_nl - ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "left", index_number_in_one_da
                )
                neutral_loss_mz_idx_max = self._find_location_from_array_with_index(
                    mz_nl + ms2_tolerance_in_da, all_nl_mass, all_nl_mass_idx_start, "right", index_number_in_one_da
                )

                # Calculate the entropy similarity for this matched peak
                modified_idx_nl = all_nl_spec_idx[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max]
                modified_value_nl = self._score_peaks_gpu(
                    entropy_transform, intensity, cp.array(all_nl_intensity[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max])
                )

                # Check if the neutral loss ion is already matched to other query peak as a product ion
                nl_matched_product_ion_mz = cp.array(ions_mz_for_nl[neutral_loss_mz_idx_min:neutral_loss_mz_idx_max])
                s1 = cp.searchsorted(product_peak_match_mz_min_gpu, nl_matched_product_ion_mz, side="right")
                s2 = cp.searchsorted(product_peak_match_mz_max_gpu, nl_matched_product_ion_mz, side="left")
                modified_value_nl[s1 > s2] = 0

                # Check if this query peak is already matched to a product ion in the same library spectrum
                duplicate_idx_in_nl = self._remove_duplicate_with_gpu(modified_idx_product, modified_idx_nl, self.total_spectra_num)
                modified_value_nl[duplicate_idx_in_nl] = 0

                entropy_similarity_modification_list.append((modified_idx_nl, modified_value_nl.get()))

            entropy_similarity = cp.zeros(self.total_spectra_num, dtype=np.float32)
            for modified_idx, modified_value in entropy_similarity_modification_list:
                entropy_similarity.scatter_add(cp.array(modified_idx), cp.array(modified_value))
            return entropy_similarity.get()
        else:
            raise ValueError("target should be cpu or gpu")

    def _remove_duplicate_with_cpu(self, array_1, array_2, max_element):
        if len(array_1) + len(array_2) < 4_000_000:
            # When len(array_1) + len(array_2) < 4_000_000, this method is faster than array method
            aux = np.concatenate((array_1, array_2))
            aux_sort_indices = np.argsort(aux, kind="mergesort")
            aux = aux[aux_sort_indices]
            mask = aux[1:] == aux[:-1]
            array2_indices = aux_sort_indices[1:][mask] - array_1.size
            return array2_indices
        else:
            # When len(array_1) + len(array_2) > 4_000_000, this method is faster than sort method
            note = np.zeros(max_element, dtype=np.int8)
            note[array_1] = 1
            duplicate_idx = np.where(note[array_2] == 1)[0]
            return duplicate_idx

    def _remove_duplicate_with_gpu(self, array_1, array_2, max_element):
        import cupy as cp

        if len(array_1) + len(array_2) < 4_000_000:
            # When len(array_1) + len(array_2) < 4_000_000, this method is faster than array method
            aux = cp.array(np.concatenate((array_1, array_2)))
            aux_sort_indices = cp.argsort(aux)
            aux = aux[aux_sort_indices]
            mask = aux[1:] == aux[:-1]
            ar2_indices = aux_sort_indices[1:][mask] - array_1.size
            return ar2_indices

        else:
            # When len(array_1) + len(array_2) > 4_000_000, this method is faster than sort method
            note = cp.zeros(max_element, dtype=np.int8)
            note[array_1] = 1
            duplicate_idx = cp.where(note[array_2] == 1)[0]
            return duplicate_idx

    def build_index(self, all_spectra_list: list, max_indexed_mz: float = 1500.00005, index_for_neutral_loss: bool = True):
        """
        Build the index for the MS/MS spectra library.

        The spectra provided to this function should be a dictionary in the format of {"precursor_mz": precursor_mz, "peaks": peaks}.
        The precursor_mz is the precursor m/z value of the MS/MS spectrum;
        The peaks is a numpy array which has been processed by the function "clean_spectrum".

        :param all_spectra_list:    A list of dictionaries in the format of {"precursor_mz": precursor_mz, "peaks": peaks},
                                    the spectra in the list need to be sorted by the precursor m/z.
        :param max_indexed_mz: The maximum m/z value that will be indexed. Default is 1500.00005.
        :param append:  Not implemented yet.
        """

        # Get the total number of spectra and peaks
        total_peaks_num = int(np.sum([spectrum["peaks"].shape[0] for spectrum in all_spectra_list]))
        total_spectra_num = len(all_spectra_list)
        self.total_spectra_num = total_spectra_num
        self.total_peaks_num = total_peaks_num

        # total_spectra_num can not be bigger than 2^32-1 (uint32), total_peak_num can not be bigger than 2^63-1 (int64)
        assert self.total_spectra_num < 2**32 - 1, "The total spectra number is too big."
        assert self.total_peaks_num < 2**63 - 1, "The total peaks number is too big."

        ############## Step 1: Collect the precursor m/z and peaks information. ##############
        peak_data = self._merge_all_spectra_to_peak_data(all_spectra_list, total_peaks_num)

        ############## Step 2: Build the index by sort with product ions. ##############
        self.index = self._generate_index_from_peak_data(peak_data, max_indexed_mz, index_for_neutral_loss=index_for_neutral_loss)
        return self.index

    def _merge_all_spectra_to_peak_data(self, all_spectra_list, total_peaks_num):
        dtype_peak_data = np.dtype(
            [
                ("ion_mz", np.float32),  # The m/z of the fragment ion.
                ("nl_mass", np.float32),  # The neutral loss mass of the fragment ion.
                ("intensity", np.float32),  # The intensity of the fragment ion.
                ("spec_idx", np.uint32),  # The index of the MS/MS spectra.
            ],
            align=True,
        )  # The index of the fragment ion.

        # Initialize the peak data array.
        peak_data = np.zeros(total_peaks_num, dtype=dtype_peak_data)
        peak_idx = 0

        # Adding the precursor m/z and peaks information to the peak data array.
        for idx, spectrum in enumerate(all_spectra_list):
            precursor_mz, peaks = spectrum["precursor_mz"], spectrum["peaks"]
            # Check the peaks array.
            assert peaks.ndim == 2, "The peaks array should be a 2D numpy array."
            assert peaks.shape[1] == 2, "The peaks array should be a 2D numpy array with the shape of [n, 2]."
            assert peaks.shape[0] > 0, "The peaks array should not be empty."
            assert abs(np.sum(peaks[:, 1]) - 1) < 1e-4, "The peaks array should be normalized to sum to 1."
            assert (
                peaks.shape[0] <= 1 or np.min(peaks[1:, 0] - peaks[:-1, 0]) > self.max_ms2_tolerance_in_da * 2
            ), "The peaks array should be sorted by m/z, and the m/z difference between two adjacent peaks should be larger than 2 * max_ms2_tolerance_in_da."

            # Preprocess the peaks array.
            peaks = self._preprocess_peaks(peaks)

            # Assign the product ion m/z
            peak_data_item = peak_data[peak_idx : (peak_idx + peaks.shape[0])]
            peak_data_item["ion_mz"] = peaks[:, 0]
            # Assign the neutral loss mass
            peak_data_item["nl_mass"] = precursor_mz - peaks[:, 0]
            # Assign the intensity
            peak_data_item["intensity"] = peaks[:, 1]
            # Assign the spectrum index
            peak_data_item["spec_idx"] = idx
            # Set the peak index
            peak_idx += peaks.shape[0]
        return peak_data

    def _generate_index(self, all_ions_mz, max_indexed_mz):
        # Build index for fast access to the ion's m/z.
        max_mz = min(np.max(all_ions_mz), max_indexed_mz)
        search_array = np.arange(0.0, max_mz, self.mz_index_step)
        mass_idx_start = np.searchsorted(all_ions_mz, search_array, side="left").astype(np.int64)
        return mass_idx_start

    def _generate_index_from_peak_data(self, peak_data, max_indexed_mz, index_for_neutral_loss):
        # Sort with precursor m/z.
        peak_data.sort(order="ion_mz")

        # Record the m/z, intensity, and spectrum index information for product ions.
        all_ions_mz = np.copy(peak_data["ion_mz"])
        all_ions_intensity = np.copy(peak_data["intensity"])
        all_ions_spec_idx = np.copy(peak_data["spec_idx"])

        # Build index for fast access to the ion's m/z.
        all_ions_mz_idx_start = self._generate_index(all_ions_mz, max_indexed_mz)
        fragment_ion_index = (
            all_ions_mz_idx_start,
            all_ions_mz,
            all_ions_intensity,
            all_ions_spec_idx,
        )

        ############## Step 3: Build the index by sort with neutral loss mass. ##############
        if index_for_neutral_loss:
            # Sort with the neutral loss mass.
            peak_data.sort(order="nl_mass")

            # Build the index for fast access to the neutral loss mass.
            all_nl_mass_idx_start = self._generate_index(peak_data["nl_mass"], max_indexed_mz)
            neutral_loss_index = (
                all_nl_mass_idx_start,
                peak_data["nl_mass"],
                peak_data["intensity"],
                peak_data["spec_idx"],
                peak_data["ion_mz"],  # The fragment ion m/z corresponding to the neutral loss mass.
            )
        else:
            neutral_loss_index = None

        return fragment_ion_index, neutral_loss_index

    def _preprocess_peaks(self, peaks):
        """
        Preprocess the peaks.
        """
        if self.intensity_weight == "entropy":
            peaks_clean = np.asarray(apply_weight_to_intensity(peaks))
            peaks_clean[:, 1] /= 2
        elif self.intensity_weight is None:
            peaks_clean = peaks.copy()
            peaks_clean[:, 1] /= 2

        return peaks_clean

    def _score_peaks_with_cpu(self, intensity_query, intensity_library):
        intensity_mix = intensity_library + intensity_query
        modified_value = intensity_mix * np.log2(intensity_mix) - intensity_library * np.log2(intensity_library) - intensity_query * np.log2(intensity_query)
        return modified_value

    def _score_peaks_gpu(self, entropy_transform, intensity_query, intensity_library):
        return entropy_transform(intensity_library, intensity_query)

    def _find_location_from_array_with_index(self, wanted_mz, mz_array, mz_idx_start_array, side, index_number):
        mz_min_int = (np.floor(wanted_mz * index_number)).astype(int)
        mz_max_int = mz_min_int + 1

        if mz_min_int >= len(mz_idx_start_array):
            mz_idx_search_start = mz_idx_start_array[-1]
        else:
            mz_idx_search_start = mz_idx_start_array[mz_min_int].astype(int)

        if mz_max_int >= len(mz_idx_start_array):
            mz_idx_search_end = len(mz_array)
        else:
            mz_idx_search_end = mz_idx_start_array[mz_max_int].astype(int) + 1

        return mz_idx_search_start + np.searchsorted(mz_array[mz_idx_search_start:mz_idx_search_end], wanted_mz, side=side)

    def save_memory_for_multiprocessing(self):
        """
        Move the numpy array in the index to shared memory in order to save memory.
        This function is not required when you only use one thread to search the MS/MS spectra.
        When use multiple threads, this function is also not required but highly recommended, as it avoids the memory copy and saves a lot of memory and time.
        """
        if self._init_for_multiprocessing:
            return
        new_index = []
        for entry in self.index:
            if entry is None:
                new_index.append(None)
                continue
            new_entry = tuple(_convert_numpy_array_to_shared_memory(a) for a in entry)
            new_index.append(new_entry)
        self.index = new_index
        self._init_for_multiprocessing = True

    def read(self, path_data=None):
        """
        Read the index from the specified path.
        """
        try:
            if path_data is None:
                path_data = self.path_data

            path_data = Path(path_data)
            with open(path_data / "information.json", "r") as f:
                information = json.load(f)
            self.mz_index_step = information["mz_index_step"]
            self.total_spectra_num = information["total_spectra_num"]
            self.total_peaks_num = information["total_peaks_num"]
            self.max_ms2_tolerance_in_da = information["max_ms2_tolerance_in_da"]

            is_nl_indexed = information["neutral_loss_indexed"]

            self.index = []
            for index_type in ["all_ions", "all_nl"]:
                if index_type == "all_nl" and not is_nl_indexed:
                    self.index.append(None)
                    continue
                library_mz_idx_start = np.memmap(path_data / f"{index_type}_mz_idx_start.bin", dtype=np.int64, mode="r")
                library_mz = np.memmap(path_data / f"{index_type}_mz.bin", dtype=np.float32, mode="r")
                library_peaks_intensity = np.memmap(path_data / f"{index_type}_intensity.bin", dtype=np.float32, mode="r")
                library_spec_idx = np.memmap(path_data / f"{index_type}_spec_idx.bin", dtype=np.uint32, mode="r")
                if index_type == "all_nl":
                    ions_mz_for_nl = np.memmap(path_data / f"ions_mz_for_nl.bin", dtype=np.float32, mode="r")
                    self.index.append((library_mz_idx_start, library_mz, library_peaks_intensity, library_spec_idx, ions_mz_for_nl))
                else:
                    self.index.append((library_mz_idx_start, library_mz, library_peaks_intensity, library_spec_idx))
            return True
        except:
            return False

    def write(self, path_data=None):
        """
        Write the index to the specified path.
        """
        if path_data is None:
            path_data = self.path_data

        path_data = Path(path_data)
        path_data.mkdir(parents=True, exist_ok=True)

        for indexes, index_type in zip(self.index, ["all_ions", "all_nl"]):
            if indexes is None:
                continue
            library_mz_idx_start, library_mz, library_peaks_intensity, library_spec_idx = indexes[:4]
            library_mz_idx_start.tofile(str(path_data / f"{index_type}_mz_idx_start.bin"))
            library_mz.tofile(str(path_data / f"{index_type}_mz.bin"))
            library_peaks_intensity.tofile(str(path_data / f"{index_type}_intensity.bin"))
            library_spec_idx.tofile(str(path_data / f"{index_type}_spec_idx.bin"))
            if index_type == "all_nl":
                ions_mz_for_nl = indexes[4]
                ions_mz_for_nl.tofile(str(path_data / f"ions_mz_for_nl.bin"))

        information = {
            "neutral_loss_indexed": self.index[1] is not None,
            "mz_index_step": float(self.mz_index_step),
            "total_spectra_num": int(self.total_spectra_num),
            "total_peaks_num": int(self.total_peaks_num),
            "max_ms2_tolerance_in_da": float(self.max_ms2_tolerance_in_da),
        }
        with open(path_data / "information.json", "w") as f:
            json.dump(information, f)


def _convert_numpy_array_to_shared_memory(np_array):
    c_type = np.ctypeslib.as_ctypes_type(np_array.dtype)
    base = multiprocessing.RawArray(c_type, int(np_array.size))
    np_array_new = np.frombuffer(base, dtype=np_array.dtype).reshape(np_array.shape)
    np_array_new[...] = np_array
    return np_array_new



#!/usr/bin/env python3
import zipfile
from pathlib import Path
from typing import Union

import numpy as np

from . import lbm2_file, mgf_file, msp_file, mzml_file


def standardize_spectrum(spectrum_dict: dict, standardize_info: dict, keep_all_keys=True):
    """
    Standardize spectrum informat to a standard format provided by standardize_info.

    :param spectrum_dict: spectrum info to standardize, this dict will be modified if keep_all_keys is True.
    :param standardize_info: {wanted_key:[[candidate_keys],default_value,default_type_function]}
    :param keep_all_keys: whether to keep all keys in spectrum_dict. If False, only output wanted keys, else output all keys.

    :return: standardized spectrum info
    """
    if keep_all_keys:
        spec_result = spectrum_dict
    else:
        spec_result = {}
    for key_target, (key_all_candidates, value_default, cast_type) in standardize_info.items():
        key_all_candidates_new = [key_target] + key_all_candidates
        for key_candidate in key_all_candidates_new:
            if key_candidate in spectrum_dict:
                try:
                    if cast_type is None:
                        spec_result[key_target] = spectrum_dict.pop(key_candidate)
                    else:
                        spec_result[key_target] = cast_type(spectrum_dict.pop(key_candidate))
                    break
                except:
                    continue
        else:
            spec_result[key_target] = value_default

        # for key_candidate in key_all_candidates:
        #     spectrum_dict.pop(key_candidate, None)
    return spec_result


def read_one_spectrum(file_input: Union[str, Path],
                      file_type: object = None,
                      **kwargs) -> dict:
    """
    A generator to read one spectrum from file.

    Currently support **.mgf**, **.msp**, **.mzML** and **.lbm2** file.

    The .mgf, .msp can be compressed with .gz, .bz2 or .zip extension.

    The .mgf format is tested with files generated by MSConvert and file downloaded from GNPS.
    The .msp format is tested with files from NIST, files downloaded from MassBank.us and GNPS, and files generated by MS-DIAL.
    The .mzML format is tested with files generated by MSConvert.
    The .lbm2 format is tested with files downloaded from MS-DIAL website.

    :param file_input: The file path of input file, can be str or Path.
    :param file_type: The file type of input file, default is None. Can be "mgf", "msp", "mzml" or "lbm2".
                        If file_type is None, the file type will be determined by file extension.
    :return: a dict contains one spectrum information. The following keys are expected:
                "_ms_level": The MS level of the spectrum, int. Expected value is 1 or 2.
                "_scan_number": The scan number of the spectrum, int. Expected value is 1 or larger.
                "peaks": The peaks of the spectrum, np.array with dtype=np.float32. Expected shape is (N, 2).
    """

    # Determine file type
    if file_type is None:
        file_type = guess_file_type_from_file_name(file_input)
    if file_type is None:
        raise ValueError("Cannot determine file type from file name: {}".format(file_input))

    if file_type == "mgf":
        spectral_generator = mgf_file.read_one_spectrum(file_input, **kwargs)
    elif file_type == "msp":
        spectral_generator = msp_file.read_one_spectrum(file_input, **kwargs)
    elif file_type == "mzml":
        spectral_generator = mzml_file.read_one_spectrum(file_input, **kwargs)
    elif file_type == "lbm2":
        spectral_generator = lbm2_file.read_one_spectrum(file_input, **kwargs)
    else:
        raise ValueError("Unknown file type: {}".format(file_type))

    for spectrum in spectral_generator:
        yield spectrum


def guess_file_type_from_file_name(filename_input):
    file_type = None
    filename_input = str(filename_input)

    # For zip file, only select the first file for the zip file
    if filename_input[-4:].lower() == ".zip":
        fzip_all = zipfile.ZipFile(filename_input)
        fzip_list = zipfile.ZipFile.namelist(fzip_all)
        # Only select the first file for the zip file
        if fzip_list[0][-4:].lower() == ".msp":
            file_type = "msp"
        elif fzip_list[0][-4:].lower() == ".mgf":
            file_type = "mgf"

    # For .gz file
    elif filename_input[-3:].lower() == ".gz":
        if filename_input[-8:-3].lower() == ".mzml":
            file_type = "mzml"
        elif filename_input[-7:-3].lower() == ".msp":
            file_type = "msp"
        elif filename_input[-7:-3].lower() == ".mgf":
            file_type = "mgf"

    # For .bz2 file
    elif filename_input[-4:].lower() == ".bz2":
        if filename_input[-8:-4].lower() == ".msp":
            file_type = "msp"
        elif filename_input[-8:-4].lower() == ".mgf":
            file_type = "mgf"

    else:
        if filename_input[-4:].lower() == ".msp":
            file_type = "msp"
        elif filename_input[-4:].lower() == ".mgf":
            file_type = "mgf"
        elif filename_input[-5:].lower() == ".mzml":
            file_type = "mzml"
        elif filename_input[-5:].lower() == ".hdf5":
            file_type = "hdf5"
        elif filename_input[-4:].lower() == ".raw":
            file_type = "raw"
        elif filename_input[-5:].lower() == ".lbm2":
            file_type = "lbm2"
        elif filename_input.split(".")[-2].lower() == "msp":
            file_type = "msp"
        elif filename_input.split(".")[-2].lower() == "mgf":
            file_type = "mgf"
        elif filename_input.split(".")[-2].lower() == "mzml":
            file_type = "mzml"

    return file_type



#!/usr/bin/env python3
import gzip
import bz2
import zipfile
import codecs


def smart_open_file(file_input, file_encoding='utf-8'):
    """
    Open file with different compression format.

    :param file_input: The file path of input file, can be str or Path.
                        The file can be compressed with .gz, .bz2 or .zip extension.    
    :param file_encoding: The encoding of input file, default is 'utf-8'.
    :return: The file object.
    """

    file_input = str(file_input)
    if file_input.lower()[-3:] == ".gz":
        fi = gzip.open(file_input)
    elif file_input.lower()[-4:] == ".bz2":
        fi = bz2.open(file_input, "rt")
    elif file_input.lower()[-4:] == ".zip":
        fzip_all = zipfile.ZipFile(file_input)
        fzip_list = zipfile.ZipFile.namelist(fzip_all)
        fi = fzip_all.open(fzip_list[0], "r")
    else:
        fi = codecs.open(file_input, "r", encoding=file_encoding, errors='ignore')
    return fi



import numpy as np
import unittest
import tempfile
from ms_entropy import FlashEntropySearch


class TestFlashEntropySearchWithCpu(unittest.TestCase):
    def setUp(self):
        spectral_library = [
            {"id": "Demo spectrum 1", "precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 1.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)},
            {
                "id": "Demo spectrum 2",
                "precursor_mz": 220.0,
                "peaks": np.array([[200.0, 1.0], [101.0, 1.0], [202.0, 1.0], [204.0, 1.0], [205.0, 1.0]], dtype=np.float32),
            },
            {
                "id": "Demo spectrum 3",
                "precursor_mz": 250.0,
                "peaks": np.array([[100.0, 1.0], [201.0, 1.0], [202.0, 1.0], [104.0, 1.0], [105.0, 1.0]], dtype=np.float32),
            },
            {"id": "Demo spectrum 4", "precursor_mz": 350.0, "peaks": [[100.0, 1.0], [101.0, 1.0], [302.0, 1.0], [104.0, 1.0], [105.0, 1.0]]},
        ]
        query_spectrum = {"precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 1.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)}

        self.flash_entropy = FlashEntropySearch()
        self.flash_entropy.build_index(spectral_library)
        query_spectrum["peaks"] = self.flash_entropy.clean_spectrum_for_search(precursor_mz=query_spectrum["precursor_mz"], peaks=query_spectrum["peaks"])
        self.query_spectrum = query_spectrum

    def test_read_and_write(self):
        path_test = tempfile.mkdtemp()
        self.flash_entropy.write(path_test)
        self.flash_entropy.read(path_test)

    def test_hybrid_search(self):
        similarity = self.flash_entropy.hybrid_search(
            precursor_mz=self.query_spectrum["precursor_mz"], peaks=self.query_spectrum["peaks"], ms2_tolerance_in_da=0.02
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.22299, 0.66897, 0.66897], decimal=5)

    def test_neutral_loss_search(self):
        similarity = self.flash_entropy.neutral_loss_search(
            precursor_mz=self.query_spectrum["precursor_mz"], peaks=self.query_spectrum["peaks"], ms2_tolerance_in_da=0.02
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.44598, 0.22299], decimal=5)
        similarity, matched_peaks = self.flash_entropy.neutral_loss_search(
            precursor_mz=self.query_spectrum["precursor_mz"],
            peaks=self.query_spectrum["peaks"],
            ms2_tolerance_in_da=0.02,
            output_matched_peak_number=True,
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.44598, 0.22299], decimal=5)
        np.testing.assert_almost_equal(matched_peaks, [4, 0, 2, 1], decimal=5)

    def test_open_search(self):
        similarity = self.flash_entropy.open_search(peaks=self.query_spectrum["peaks"], ms2_tolerance_in_da=0.02)
        np.testing.assert_almost_equal(similarity, [1.0, 0.22299, 0.22299, 0.44598], decimal=5)
        similarity, matched_peaks = self.flash_entropy.open_search(
            peaks=self.query_spectrum["peaks"], ms2_tolerance_in_da=0.02, output_matched_peak_number=True
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.22299, 0.22299, 0.44598], decimal=5)
        np.testing.assert_almost_equal(matched_peaks, [4, 1, 1, 2], decimal=5)

    def test_identity_search(self):
        similarity = self.flash_entropy.identity_search(
            precursor_mz=self.query_spectrum["precursor_mz"], peaks=self.query_spectrum["peaks"], ms1_tolerance_in_da=0.01, ms2_tolerance_in_da=0.02
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.0, 0.0], decimal=5)
        similarity, matched_peaks = self.flash_entropy.identity_search(
            precursor_mz=self.query_spectrum["precursor_mz"],
            peaks=self.query_spectrum["peaks"],
            ms1_tolerance_in_da=0.01,
            ms2_tolerance_in_da=0.02,
            output_matched_peak_number=True,
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.0, 0.0], decimal=5)
        np.testing.assert_almost_equal(matched_peaks, [4, 0, 0, 0], decimal=5)


class TestUnweightedFlashEntropySearchWithCpu(unittest.TestCase):
    def setUp(self):
        spectral_library = [
            {"id": "Demo spectrum 1", "precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 10.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)},
            {
                "id": "Demo spectrum 2",
                "precursor_mz": 220.0,
                "peaks": np.array([[200.0, 1.0], [101.0, 1.0], [202.0, 1.0], [204.0, 1.0], [205.0, 1.0]], dtype=np.float32),
            },
            {
                "id": "Demo spectrum 3",
                "precursor_mz": 250.0,
                "peaks": np.array([[100.0, 1.0], [201.0, 1.0], [202.0, 1.0], [104.0, 1.0], [105.0, 1.0]], dtype=np.float32),
            },
            {"id": "Demo spectrum 4", "precursor_mz": 350.0, "peaks": [[100.0, 1.0], [101.0, 1.0], [302.0, 1.0], [104.0, 1.0], [105.0, 1.0]]},
        ]
        query_spectrum = {"precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 10.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)}

        self.flash_entropy = FlashEntropySearch(intensity_weight=None)
        self.flash_entropy.build_index(spectral_library)
        query_spectrum["peaks"] = self.flash_entropy.clean_spectrum_for_search(precursor_mz=query_spectrum["precursor_mz"], peaks=query_spectrum["peaks"])
        self.query_spectrum = query_spectrum

    def test_open_search(self):
        similarity = self.flash_entropy.open_search(peaks=self.query_spectrum["peaks"], ms2_tolerance_in_da=0.02)
        np.testing.assert_almost_equal(similarity, [1.0, 0.356, 0.118, 0.474], decimal=3)
        similarity, matched_peaks = self.flash_entropy.open_search(
            peaks=self.query_spectrum["peaks"], ms2_tolerance_in_da=0.02, output_matched_peak_number=True
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.356, 0.118, 0.474], decimal=3)
        np.testing.assert_almost_equal(matched_peaks, [4, 1, 1, 2], decimal=5)

    def test_identity_search(self):
        similarity = self.flash_entropy.identity_search(
            precursor_mz=self.query_spectrum["precursor_mz"], peaks=self.query_spectrum["peaks"], ms1_tolerance_in_da=0.01, ms2_tolerance_in_da=0.02
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.0, 0.0], decimal=5)
        similarity, matched_peaks = self.flash_entropy.identity_search(
            precursor_mz=self.query_spectrum["precursor_mz"],
            peaks=self.query_spectrum["peaks"],
            ms1_tolerance_in_da=0.01,
            ms2_tolerance_in_da=0.02,
            output_matched_peak_number=True,
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.0, 0.0], decimal=5)
        np.testing.assert_almost_equal(matched_peaks, [4, 0, 0, 0], decimal=5)


class TestFlashEntropySearchWithCpuLowMemory(TestFlashEntropySearchWithCpu):
    def setUp(self):
        spectral_library = [
            {"id": "Demo spectrum 1", "precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 1.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)},
            {
                "id": "Demo spectrum 2",
                "precursor_mz": 220.0,
                "peaks": np.array([[200.0, 1.0], [101.0, 1.0], [202.0, 1.0], [204.0, 1.0], [205.0, 1.0]], dtype=np.float32),
            },
            {
                "id": "Demo spectrum 3",
                "precursor_mz": 250.0,
                "peaks": np.array([[100.0, 1.0], [201.0, 1.0], [202.0, 1.0], [104.0, 1.0], [105.0, 1.0]], dtype=np.float32),
            },
            {"id": "Demo spectrum 4", "precursor_mz": 350.0, "peaks": [[100.0, 1.0], [101.0, 1.0], [302.0, 1.0], [104.0, 1.0], [105.0, 1.0]]},
        ]
        query_spectrum = {"precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 1.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)}

        self.flash_entropy = FlashEntropySearch()
        self.flash_entropy.build_index(spectral_library)
        path_test = tempfile.mkdtemp()
        self.flash_entropy.write(path_test)
        self.flash_entropy = FlashEntropySearch(low_memory=1)
        self.flash_entropy.read(path_test)
        query_spectrum["peaks"] = self.flash_entropy.clean_spectrum_for_search(precursor_mz=query_spectrum["precursor_mz"], peaks=query_spectrum["peaks"])
        self.query_spectrum = query_spectrum

    def test_read_and_write(self):
        pass


class TestFlashEntropySearchWithCpuMediumMemory(TestFlashEntropySearchWithCpu):
    def setUp(self):
        spectral_library = [
            {"id": "Demo spectrum 1", "precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 1.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)},
            {
                "id": "Demo spectrum 2",
                "precursor_mz": 220.0,
                "peaks": np.array([[200.0, 1.0], [101.0, 1.0], [202.0, 1.0], [204.0, 1.0], [205.0, 1.0]], dtype=np.float32),
            },
            {
                "id": "Demo spectrum 3",
                "precursor_mz": 250.0,
                "peaks": np.array([[100.0, 1.0], [201.0, 1.0], [202.0, 1.0], [104.0, 1.0], [105.0, 1.0]], dtype=np.float32),
            },
            {"id": "Demo spectrum 4", "precursor_mz": 350.0, "peaks": [[100.0, 1.0], [101.0, 1.0], [302.0, 1.0], [104.0, 1.0], [105.0, 1.0]]},
        ]
        query_spectrum = {"precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 1.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)}

        self.flash_entropy = FlashEntropySearch()
        self.flash_entropy.build_index(spectral_library)
        path_test = tempfile.mkdtemp()
        self.flash_entropy.write(path_test)
        self.flash_entropy = FlashEntropySearch(low_memory=2)
        self.flash_entropy.read(path_test)
        query_spectrum["peaks"] = self.flash_entropy.clean_spectrum_for_search(precursor_mz=query_spectrum["precursor_mz"], peaks=query_spectrum["peaks"])
        self.query_spectrum = query_spectrum

    def test_read_and_write(self):
        pass


# class TestFlashEntropySearchWithGpu(TestFlashEntropySearchWithCpu):
#     def test_hybrid_search(self):
#         similarity = self.flash_entropy.hybrid_search(precursor_mz=self.query_spectrum['precursor_mz'],
#                                                       peaks=self.query_spectrum['peaks'],
#                                                       ms2_tolerance_in_da=0.02, target='gpu')
#         np.testing.assert_almost_equal(similarity, [1.0, 0.22299, 0.66897, 0.66897], decimal=5)

#     def test_neutral_loss_search(self):
#         similarity = self.flash_entropy.neutral_loss_search(precursor_mz=self.query_spectrum['precursor_mz'],
#                                                             peaks=self.query_spectrum['peaks'],
#                                                             ms2_tolerance_in_da=0.02, target='gpu')
#         np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.44598, 0.22299], decimal=5)

#     def test_open_search(self):
#         similarity = self.flash_entropy.open_search(peaks=self.query_spectrum['peaks'], ms2_tolerance_in_da=0.02, target='gpu')
#         np.testing.assert_almost_equal(similarity, [1.0, 0.22299, 0.22299, 0.44598], decimal=5)

#     def test_identity_search(self):
#         similarity = self.flash_entropy.identity_search(precursor_mz=self.query_spectrum['precursor_mz'],
#                                                         peaks=self.query_spectrum['peaks'],
#                                                         ms1_tolerance_in_da=0.01, ms2_tolerance_in_da=0.02, target='gpu')
#         np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.0, 0.0], decimal=5)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from typing import Union
from . import tools


def calculate_entropy_similarity(
    peaks_a,
    peaks_b,
    ms2_tolerance_in_da: float = 0.02,
    ms2_tolerance_in_ppm: float = -1,
    clean_spectra: bool = True,
    **kwargs
):
    """Calculate the entropy similarity between two spectra.

    First, the spectra are cleaned by the `clean_spectrum()` function.

    Then, the entropy based intensity weights are applied to the peaks.

    Finally, the entropy similarity is calculated by the `calculate_unweighted_entropy_similarity()` function.

    The formula for entropy similarity is as follows:

    .. math::
        Similarity = \\frac{1}{2} \\sum_{i,j} \\begin{cases}
        0 & \\text{ if } mz_{A,i} \\neq mz_{B,j} \\\\
        f(I_{A,i}+I_{B,j}) - f(I_{A,i}) - f(I_{B,j}) & \\text{ if } mz_{A,i} = mz_{B,j}
        \\end{cases}

    .. math::
        \\text{ where } f(x) = x \\log_2(x) \\text{ and } \\sum_{i} I_{A,i} = \\sum_{j} I_{B,j} = 1

    Parameters
    ----------
    peaks_a : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The first spectrum to calculate entropy similarity for. The first column is m/z, and the second column is intensity.

    peaks_b : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The second spectrum to calculate entropy similarity for. The first column is m/z, and the second column is intensity.

    ms2_tolerance_in_da : float, optional
        The MS2 tolerance in Da. Defaults to 0.02. If this is set to a negative value, ms2_tolerance_in_ppm will be used instead.

    ms2_tolerance_in_ppm : float, optional
        The MS2 tolerance in ppm. Defaults to -1. If this is set to a negative value, ms2_tolerance_in_da will be used instead.

        **Note:** Either `ms2_tolerance_in_da` or `ms2_tolerance_in_ppm` must be positive. If both `ms2_tolerance_in_da` and `ms2_tolerance_in_ppm` are positive, `ms2_tolerance_in_ppm` will be used.

    clean_spectra : bool, optional
        Whether to clean the spectra before calculating entropy similarity. Defaults to True. **Only set this to False if the spectra have been preprocessed by the `clean_spectrum()` function!** Otherwise, the results will be incorrect. If the spectra are already cleaned, set this to False to save time.

    **kwargs : optional
        The arguments and keyword arguments to pass to function ``clean_spectrum()``.

        _

    Returns
    -------
    float
        The entropy similarity between the two spectra.
    """
    if clean_spectra:
        kwargs.update(
            {
                "min_ms2_difference_in_da": max(2 * ms2_tolerance_in_da, kwargs.get("min_ms2_difference_in_da", -1)),
                "min_ms2_difference_in_ppm": max(2 * ms2_tolerance_in_ppm, kwargs.get("min_ms2_difference_in_ppm", -1)),
            }
        )
        peaks_a = tools.clean_spectrum(peaks_a, **kwargs)
        peaks_b = tools.clean_spectrum(peaks_b, **kwargs)
    else:
        peaks_a = np.asarray(peaks_a, dtype=np.float32, order="C").reshape(-1, 2)
        peaks_b = np.asarray(peaks_b, dtype=np.float32, order="C").reshape(-1, 2)

    # Apply the weights to the peaks.
    peaks_a = apply_weight_to_intensity(peaks_a)
    peaks_b = apply_weight_to_intensity(peaks_b)

    return calculate_unweighted_entropy_similarity(
        peaks_a, peaks_b, ms2_tolerance_in_da=ms2_tolerance_in_da, ms2_tolerance_in_ppm=ms2_tolerance_in_ppm, clean_spectra=False
    )


def calculate_unweighted_entropy_similarity(
    peaks_a,
    peaks_b,
    ms2_tolerance_in_da: float = 0.02,
    ms2_tolerance_in_ppm: float = -1,
    clean_spectra: bool = True,
    **kwargs
):
    """Calculate the unweighted entropy similarity between two spectra.

    The formula for unweighted entropy similarity is as follows:

    .. math::
        Similarity = \\frac{1}{2} \\sum_{i,j} \\begin{cases}
        0 & \\text{ if } mz_{A,i} \\neq mz_{B,j} \\\\
        f(I_{A,i}+I_{B,j}) - f(I_{A,i}) - f(I_{B,j}) & \\text{ if } mz_{A,i} = mz_{B,j}
        \\end{cases}

    .. math::
        \\text{ where } f(x) = x \\log_2(x) \\text{ and } \\sum_{i} I_{A,i} = \\sum_{j} I_{B,j} = 1


    Parameters
    ----------
    peaks_a : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The first spectrum to calculate unweighted entropy similarity for. The first column is m/z, and the second column is intensity.

    peaks_b : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The second spectrum to calculate unweighted entropy similarity for. The first column is m/z, and the second column is intensity.

    ms2_tolerance_in_da : float, optional
        The MS2 tolerance in Da. Defaults to 0.02. If this is set to a negative value, ms2_tolerance_in_ppm will be used instead.

    ms2_tolerance_in_ppm : float, optional
        The MS2 tolerance in ppm. Defaults to -1. If this is set to a negative value, ms2_tolerance_in_da will be used instead.

        **Note:** Either `ms2_tolerance_in_da` or `ms2_tolerance_in_ppm` must be positive. If both `ms2_tolerance_in_da` and `ms2_tolerance_in_ppm` are positive, `ms2_tolerance_in_ppm` will be used.

    clean_spectra : bool, optional
        Whether to clean the spectra before calculating unweighted entropy similarity. Defaults to True. Only set this to False if the spectra have been preprocessed by the clean_spectrum() function! Otherwise, the results will be incorrect. If the spectra are already cleaned, set this to False to save time. If the spectra are in the list format, always set this to True or an error will be raised.

    **kwargs : optional
        The arguments and keyword arguments to pass to function ``clean_spectrum()``.

        _

    Returns
    -------
    float
        The unweighted entropy similarity between the two spectra.
    """
    if clean_spectra:
        kwargs.update(
            {
                "min_ms2_difference_in_da": max(2 * ms2_tolerance_in_da, kwargs.get("min_ms2_difference_in_da", -1)),
                "min_ms2_difference_in_ppm": max(2 * ms2_tolerance_in_ppm, kwargs.get("min_ms2_difference_in_ppm", -1)),
            }
        )
        peaks_a = tools.clean_spectrum(peaks_a, **kwargs)
        peaks_b = tools.clean_spectrum(peaks_b, **kwargs)
    else:
        peaks_a = np.asarray(peaks_a, dtype=np.float32, order="C").reshape(-1, 2)
        peaks_b = np.asarray(peaks_b, dtype=np.float32, order="C").reshape(-1, 2)

    if peaks_a.shape[0] == 0 or peaks_b.shape[0] == 0:
        return 0.0

    # Calculate the entropy similarity of the two spectra.
    a: int = 0
    b: int = 0
    peak_a_intensity: float = 0.0
    peak_b_intensity: float = 0.0
    peak_ab_intensity: float = 0.0
    entropy_similarity: float = 0.0

    max_allowed_mass_difference: float = ms2_tolerance_in_da

    while a < peaks_a.shape[0] and b < peaks_b.shape[0]:
        mass_difference: float = peaks_a[a, 0] - peaks_b[b, 0]
        if ms2_tolerance_in_ppm > 0:
            max_allowed_mass_difference = peaks_a[a, 0] * ms2_tolerance_in_ppm * 1e-6
        if mass_difference < -max_allowed_mass_difference:
            # This peak only exists in peaks_a.
            a += 1
        elif mass_difference > max_allowed_mass_difference:
            # This peak only exists in peaks_b.
            b += 1
        else:
            # This peak exists in both peaks_a and peaks_b.
            peak_a_intensity = peaks_a[a, 1]
            peak_b_intensity = peaks_b[b, 1]
            peak_ab_intensity = peak_a_intensity + peak_b_intensity
            entropy_similarity += (
                peak_ab_intensity * np.log2(peak_ab_intensity) - peak_a_intensity * np.log2(peak_a_intensity) - peak_b_intensity * np.log2(peak_b_intensity)
            )
            a += 1
            b += 1

    entropy_similarity /= 2
    return entropy_similarity


def apply_weight_to_intensity(peaks: np.ndarray) -> np.ndarray:
    """
    Apply a weight to the intensity of a spectrum based on spectral entropy based on the method described in:

    Li, Y., Kind, T., Folz, J. et al. Spectral entropy outperforms MS/MS dot product similarity for small-molecule compound identification. Nat Methods 18, 1524\\-1531 (2021). https://doi.org/10.1038/s41592-021-01331-z.

    Parameters
    ----------
    peaks : np.ndarray in shape (n_peaks, 2), np.float32
        The spectrum to apply weight to. The first column is m/z, and the second column is intensity.
        The peaks need to be pre-cleaned.

        _

    Returns
    -------
    np.ndarray in shape (n_peaks, 2), np.float32
        The spectrum with weight applied. The first column is m/z, and the second column is intensity.
        The peaks will be a copy of the input peaks.
    """
    if peaks.shape[0] == 0:
        return np.empty((0, 2), dtype=np.float32)

    # Calculate the spectral entropy.
    entropy = 0.0
    if peaks.shape[0] > 0:
        entropy = -np.sum(peaks[:, 1] * np.log(peaks[:, 1]))

    # Copy the peaks.
    weighted_peaks = peaks.copy()

    # Apply the weight.
    if entropy < 3:
        weight = 0.25 + 0.25 * entropy
        weighted_peaks[:, 1] = np.power(peaks[:, 1], weight)
        intensity_sum = np.sum(weighted_peaks[:, 1])
        weighted_peaks[:, 1] /= intensity_sum

    return weighted_peaks


def calculate_spectral_entropy(peaks, clean_spectrum=True, **kwargs) -> float:
    """Calculate the spectral entropy of a spectrum.

    Parameters
    ----------
    peaks : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The spectrum to calculate spectral entropy for. The first column is m/z, and the second column is intensity.

    clean_spectrum : bool, optional
        Whether to clean the spectrum before calculating spectral entropy. Defaults to True. If the spectrum is already cleaned, set this to False to save time.

    **kwargs : optional
        The arguments and keyword arguments to pass to clean_spectrum().

        _

    Returns
    -------
    float
        The spectral entropy of the spectrum.
    """
    # Clean the spectrum.
    if clean_spectrum:
        peaks = tools.clean_spectrum(peaks, **kwargs)
    else:
        peaks = np.asarray(peaks, dtype=np.float32, order="C").reshape((-1, 2))

    # Calculate the spectral entropy.
    if peaks.shape[0] == 0:
        return 0.0
    else:
        spectral_entropy = -np.sum(peaks[:, 1] * np.log(peaks[:, 1]))
        return spectral_entropy



from .spectra import (
    calculate_spectral_entropy,
    clean_spectrum,
    calculate_entropy_similarity,
    calculate_unweighted_entropy_similarity,
    apply_weight_to_intensity,
)
from .file_io import read_one_spectrum, standardize_spectrum
from .entropy_search import (
    FlashEntropySearch,
    FlashEntropySearchCore,
    FlashEntropySearchCoreLowMemory,
    FlashEntropySearchCoreMediumMemory,
    FlashEntropySearchCoreForDynamicIndexing,
)
from .version import __version__



import numpy as np
import unittest
import tempfile
from ms_entropy import FlashEntropySearchCoreForDynamicIndexing, clean_spectrum


class TestFlashEntropySearchWithCpu(unittest.TestCase):
    def setUp(self):
        path_test = tempfile.mkdtemp()
        spectral_library = [
            {"id": "Demo spectrum 1", "precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 1.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)},
            {
                "id": "Demo spectrum 2",
                "precursor_mz": 220.0,
                "peaks": np.array([[200.0, 1.0], [101.0, 1.0], [202.0, 1.0], [204.0, 1.0], [205.0, 1.0]], dtype=np.float32),
            },
            {
                "id": "Demo spectrum 3",
                "precursor_mz": 250.0,
                "peaks": np.array([[100.0, 1.0], [201.0, 1.0], [202.0, 1.0], [104.0, 1.0], [105.0, 1.0]], dtype=np.float32),
            },
            {"id": "Demo spectrum 4", "precursor_mz": 350.0, "peaks": [[100.0, 1.0], [101.0, 1.0], [302.0, 1.0], [104.0, 1.0], [105.0, 1.0]]},
        ]
        query_spectrum = {"precursor_mz": 150.0, "peaks": np.array([[100.0, 1.0], [101.0, 1.0], [102.0, 1.0], [103.0, 1.0]], dtype=np.float32)}

        for spec in spectral_library:
            # Clean the peaks
            spec["peaks"] = clean_spectrum(
                peaks=spec["peaks"],
                max_mz=spec["precursor_mz"],
                noise_threshold=0.01,
                min_ms2_difference_in_da=0.02,
            )

        self.flash_entropy = FlashEntropySearchCoreForDynamicIndexing(path_data=path_test)
        self.flash_entropy.build_index(spectral_library)
        query_spectrum["peaks"] = clean_spectrum(max_mz=query_spectrum["precursor_mz"], peaks=query_spectrum["peaks"])
        self.query_spectrum = query_spectrum
        self.spectral_library = spectral_library

    def test_read_and_write(self):
        self.flash_entropy.write()
        self.flash_entropy.read()

    def test_hybrid_search(self):
        similarity = self.flash_entropy.search_hybrid(
            precursor_mz=self.query_spectrum["precursor_mz"], peaks=self.query_spectrum["peaks"], ms2_tolerance_in_da=0.02
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.22299, 0.66897, 0.66897], decimal=5)

    def test_neutral_loss_search(self):
        similarity = self.flash_entropy.search(
            precursor_mz=self.query_spectrum["precursor_mz"], peaks=self.query_spectrum["peaks"], method="neutral_loss", ms2_tolerance_in_da=0.02
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.44598, 0.22299], decimal=5)
        similarity, matched_peaks = self.flash_entropy.search(
            precursor_mz=self.query_spectrum["precursor_mz"],
            peaks=self.query_spectrum["peaks"],
            method="neutral_loss",
            ms2_tolerance_in_da=0.02,
            output_matched_peak_number=True,
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.44598, 0.22299], decimal=5)
        np.testing.assert_almost_equal(matched_peaks, [4, 0, 2, 1], decimal=5)

    def test_open_search(self):
        similarity = self.flash_entropy.search(peaks=self.query_spectrum["peaks"], method="open", ms2_tolerance_in_da=0.02)
        np.testing.assert_almost_equal(similarity, [1.0, 0.22299, 0.22299, 0.44598], decimal=5)
        similarity, matched_peaks = self.flash_entropy.search(
            peaks=self.query_spectrum["peaks"], method="open", ms2_tolerance_in_da=0.02, output_matched_peak_number=True
        )
        np.testing.assert_almost_equal(similarity, [1.0, 0.22299, 0.22299, 0.44598], decimal=5)
        np.testing.assert_almost_equal(matched_peaks, [4, 1, 1, 2], decimal=5)

    def test_identity_search(self):
        similarity = self.flash_entropy.search(method="open", peaks=self.query_spectrum["peaks"], ms2_tolerance_in_da=0.02)
        library_precursor_mz = np.array([x["precursor_mz"] for x in self.spectral_library], dtype=np.float32)
        similarity[np.abs(library_precursor_mz - self.query_spectrum["precursor_mz"]) > 0.01] = 0
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.0, 0.0], decimal=5)
        similarity, matched_peaks = self.flash_entropy.search(
            method="open",
            peaks=self.query_spectrum["peaks"],
            ms2_tolerance_in_da=0.02,
            output_matched_peak_number=True,
        )
        similarity[np.abs(library_precursor_mz - self.query_spectrum["precursor_mz"]) > 0.01] = 0
        matched_peaks[np.abs(library_precursor_mz - self.query_spectrum["precursor_mz"]) > 0.01] = 0
        np.testing.assert_almost_equal(similarity, [1.0, 0.0, 0.0, 0.0], decimal=5)
        np.testing.assert_almost_equal(matched_peaks, [4, 0, 0, 0], decimal=5)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import numpy as np


def read_one_spectrum(file_input, **kwargs) -> dict:
    """
    A generator to read one spectrum from .mzml file.

    :param filename_input: The file path of input file, can be str or Path.
    :param obo_version: The version of obo file, default is "4.1.33".
    :return: a dict contains one spectrum information.
    """
    import pyteomics.mzml
    with pyteomics.mzml.read(str(file_input)) as reader:
        for scan, spec in enumerate(reader):
            mz = spec['m/z array']
            intensity = spec['intensity array']
            peaks = np.asarray([mz, intensity], dtype=np.float32).T
            try:
                rt = spec['scanList']['scan'][0]['scan start time']
            except:
                rt = -1

            try:
                precursor_mz = spec['precursorList']['precursor'][0]['selectedIonList']['selectedIon'][0]['selected ion m/z']
            except:
                precursor_mz = -1

            try:
                charge = spec['precursorList']['precursor'][0]['selectedIonList']['selectedIon'][0]['charge state']
            except:
                charge = 1

            if 'negative scan' in spec:
                charge = -1 * np.abs(charge)
            elif 'positive scan' in spec:
                charge = np.abs(charge)

            spectrum_info = {
                "_ms_level": spec.get("ms level", 0),
                "_scan_number": scan + 1,
                "peaks": np.asarray(peaks, dtype=np.float32, order="C"),
                "rt": rt,
                "charge": charge,
                "precursor_mz": precursor_mz,
                "name": spec.get("spectrum title", ""),
            }
            yield spectrum_info


def read_one_spectrum_with_pymzml(file_input, **kwargs) -> dict:
    """
    A generator to read one spectrum from .mzml file.

    :param filename_input: The file path of input file, can be str or Path.
    :param obo_version: The version of obo file, default is "4.1.33".
    :return: a dict contains one spectrum information.
    """
    import pymzml
    run = pymzml.run.Reader(file_input, obo_version="4.1.33")
    for _scan_number, raw_spectrum_info in enumerate(run):
        spectrum_info = {
            "_ms_level": raw_spectrum_info.ms_level,
            "_scan_number": _scan_number + 1,
            "peaks": np.asarray(raw_spectrum_info.peaks("centroided"), dtype=np.float32, order="C"),
            'rt': raw_spectrum_info.scan_time_in_minutes() * 60,
            'precursor_mz': raw_spectrum_info.selected_precursors[0].get('mz', None) if len(raw_spectrum_info.selected_precursors) > 0 else None,
            'precursor_charge': raw_spectrum_info.selected_precursors[0].get('charge', None) if len(raw_spectrum_info.selected_precursors) > 0 else None,
        }
        try:
            spectrum_title = raw_spectrum_info.get_element_by_name("spectrum title")
            spectrum_info["name"] = spectrum_title.attrib["value"]
        except:
            spectrum_info["name"] = ""

        if spectrum_info["precursor_charge"] is None:
            if raw_spectrum_info["negative scan"]:
                spectrum_info["precursor_charge"] = -1
            elif raw_spectrum_info["positive scan"]:
                spectrum_info["precursor_charge"] = 1
        else:
            try:
                spectrum_info["precursor_charge"] = int(spectrum_info["precursor_charge"])
                if raw_spectrum_info["negative scan"]:
                    spectrum_info["precursor_charge"] = -abs(spectrum_info["precursor_charge"])
                elif raw_spectrum_info["positive scan"]:
                    spectrum_info["precursor_charge"] = abs(spectrum_info["precursor_charge"])
            except ValueError:
                print("Warning: precursor charge is not an integer: {}".format(spectrum_info["precursor_charge"]))

        yield spectrum_info

    run.close()



import numpy as np
import unittest
import ms_entropy as me


class TestSpectralEntropy(unittest.TestCase):
    def test_clean_spectrum(self):
        # Test
        peaks = np.array([[41.04, 0.3716], [69.071, 7.917962], [69.071, 100.0], [86.0969, 66.83]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, max_mz=85, noise_threshold=0.01, min_ms2_difference_in_da=0.05)
        np.testing.assert_allclose(peaks, np.array([[69.071, 1.0]], dtype=np.float32), atol=1e-4)

        print("*" * 80)
        print("Test start")
        # Test
        peaks = np.array([[41.04, 0.3716], [69.070, 0.917962], [69.071, 100.0], [86.0969, 66.83]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, noise_threshold=0.01, min_ms2_difference_in_ppm=50)
        np.testing.assert_allclose(peaks, np.array([[69.07099, 0.6016047], [86.0969, 0.39839533]], dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([[41.04, 0.3716], [69.070, 0.917962], [69.071, 100.0], [86.0969, 66.83]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, max_mz=85, noise_threshold=0.01, min_ms2_difference_in_ppm=50)
        np.testing.assert_allclose(peaks, np.array([[69.07099, 1.0]], dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([[41.04, 1.3716], [69.071, 100.0], [86.0969, 66.83]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, max_mz=85, noise_threshold=0.01, min_ms2_difference_in_ppm=50, normalize_intensity=False)
        np.testing.assert_allclose(peaks, np.array([[41.04, 1.3716], [69.071, 100.0]], dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, max_mz=85, noise_threshold=0.01, min_ms2_difference_in_da=0.05)
        np.testing.assert_allclose(peaks, np.zeros((0, 2), dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([[41.04, 0.0], [69.071, 0.0], [86.0969, 66.83]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, max_mz=85, noise_threshold=0.01, min_ms2_difference_in_da=0.05, normalize_intensity=False)
        np.testing.assert_allclose(peaks, np.zeros((0, 2), dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([[86.0969, 66.83]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, max_mz=85, noise_threshold=0.01, min_ms2_difference_in_da=0.05)
        np.testing.assert_allclose(peaks, np.zeros((0, 2), dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([[41.04, 20.3716], [69.071, 7.917962], [86.0969, 66.83]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, max_mz=85, min_mz=50, noise_threshold=0.01, min_ms2_difference_in_da=0.05)
        np.testing.assert_allclose(peaks, np.array([[69.071, 1.0]], dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([[41.04, 10], [69.071, 30], [86.0969, 60]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, max_mz=-1, min_mz=-1, noise_threshold=0.01, min_ms2_difference_in_da=0.05)
        np.testing.assert_allclose(peaks, np.array([[41.04, 0.1], [69.071, 0.3], [86.0969, 0.6]], dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([[41.04, 10], [69.071, 40], [86.0969, 60]], dtype=np.float32)
        peaks = me.clean_spectrum(peaks, noise_threshold=0.01, min_ms2_difference_in_da=0.05, max_peak_num=2)
        np.testing.assert_allclose(peaks, np.array([[69.071, 0.4], [86.0969, 0.6]], dtype=np.float32), atol=1e-4)

        # Test
        peaks = np.array([[86.0969, 66.83]], dtype=np.float32)
        self.assertRaises(ValueError, me.clean_spectrum, peaks, noise_threshold=0.01, min_ms2_difference_in_da=-1)

    def test_spectral_entropy(self):
        # Test
        peaks = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        peaks = me.clean_spectrum(
            peaks,
            min_mz=0.0,
            max_mz=1000.0,
            noise_threshold=0.01,
            min_ms2_difference_in_da=0.05,
            max_peak_num=-1,
            normalize_intensity=True,
        )
        entropy = me.calculate_spectral_entropy(peaks, clean_spectrum=False)
        self.assertAlmostEqual(entropy, 0.2605222463607788, places=4)

        # Test
        peaks = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        entropy = me.calculate_spectral_entropy(
            peaks, clean_spectrum=True, min_mz=0, max_mz=1000, noise_threshold=0.01, min_ms2_difference_in_da=0.05, max_peak_num=None, normalize_intensity=True
        )
        self.assertAlmostEqual(entropy, 0.2605222463607788, places=4)

        # Test
        peaks = np.array([[86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        entropy = me.calculate_spectral_entropy(
            peaks, clean_spectrum=True, min_mz=0, max_mz=85, noise_threshold=0.01, min_ms2_difference_in_da=0.05, max_peak_num=None, normalize_intensity=True
        )
        self.assertAlmostEqual(entropy, 0.0, places=4)


class TestEntropySimilarity(unittest.TestCase):
    def test_unweighted_entropy_similarity(self):
        # Test
        spec_query = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        spec_reference = np.array([[41.04, 37.16], [69.07, 66.83], [86.1, 999.0]], dtype=np.float32)
        similarity = me.calculate_unweighted_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_da=0.05)
        self.assertAlmostEqual(similarity, 0.9826668790176113, places=4)

        # Test
        similarity = me.calculate_unweighted_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_da=0.015)
        self.assertAlmostEqual(similarity, 0.9780402849658875, places=4)

        # Test
        spec_query = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        spec_reference = np.array([[41.04, 37.16], [69.07, 66.83], [86.1, 999.0]], dtype=np.float32)
        similarity = me.calculate_unweighted_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_da=0.05, max_mz=50)
        self.assertAlmostEqual(similarity, 0.0, places=4)

    def test_unweighted_entropy_similarity_ppm(self):
        spec_query = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        spec_reference = np.array([[41.04, 37.16], [69.07, 66.83], [86.1, 999.0]], dtype=np.float32)
        similarity = me.calculate_unweighted_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_ppm=200)
        self.assertAlmostEqual(similarity, 0.9826668790176113, places=4)

        similarity = me.calculate_unweighted_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_ppm=50)
        self.assertAlmostEqual(similarity, 0.9780402849658875, places=4)

    def test_entropy_similarity(self):
        spec_query = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        spec_reference = np.array([[41.04, 37.16], [69.07, 66.83], [86.1, 999.0]], dtype=np.float32)
        similarity = me.calculate_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_da=0.05)
        self.assertAlmostEqual(similarity, 0.8984397722577456, places=4)

        similarity = me.calculate_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_da=0.015)
        self.assertAlmostEqual(similarity, 0.8380874603833756, places=4)

        # Test
        spec_query = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        spec_reference = np.array([[41.04, 37.16], [69.07, 66.83], [86.1, 999.0]], dtype=np.float32)
        similarity = me.calculate_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_da=0.05, max_mz=50)
        self.assertAlmostEqual(similarity, 0.0, places=4)

        # Test
        spec_query = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        spec_reference = np.array([[41.04, 37.16], [69.07, 66.83], [86.1, 999.0]], dtype=np.float32)
        spec_query = me.clean_spectrum(
            spec_query,
            min_mz=0,
            max_mz=85,
            noise_threshold=0.01,
        )
        spec_reference = me.clean_spectrum(
            spec_reference,
            min_mz=50,
            max_mz=85,
            noise_threshold=0.01,
        )
        similarity = me.calculate_entropy_similarity(spec_query, spec_reference, clean_spectra=False)
        self.assertAlmostEqual(similarity, 1.0, places=4)

    def test_entropy_similarity_ppm(self):
        spec_query = np.array([[69.071, 7.917962], [86.066, 1.021589], [86.0969, 100.0]], dtype=np.float32)
        spec_reference = np.array([[41.04, 37.16], [69.07, 66.83], [86.1, 999.0]], dtype=np.float32)
        similarity = me.calculate_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_ppm=200)
        self.assertAlmostEqual(similarity, 0.8984397722577456, places=4)

        similarity = me.calculate_entropy_similarity(spec_query, spec_reference, ms2_tolerance_in_ppm=50)
        self.assertAlmostEqual(similarity, 0.8380874603833756, places=4)


if __name__ == "__main__":
    unittest.main()



from .entropy import calculate_spectral_entropy, apply_weight_to_intensity

try:
    from .entropy_cython import (
        cy_clean_spectrum as clean_spectrum,
        cy_calculate_entropy_similarity as calculate_entropy_similarity,
        cy_calculate_unweighted_entropy_similarity as calculate_unweighted_entropy_similarity,
    )
except ImportError:
    from .tools import clean_spectrum
    from .entropy import calculate_entropy_similarity, calculate_unweighted_entropy_similarity



import numpy as np
cimport numpy as np

ctypedef np.float32_t float32
ctypedef np.int64_t int_64
ctypedef np.int8_t int_8
ctypedef np.uint32_t uint_32
ctypedef np.int32_t int_32

ctypedef np.float32_t float_spec
ctypedef int bool

cdef extern from "CleanSpectrum.h":
    int clean_spectrum(float_spec* peaks, int peaks_length,
                        float min_mz, float max_mz,
                        float noise_threshold,
                        float min_ms2_difference_in_da, float min_ms2_difference_in_ppm,
                        int max_peak_num,
                        int normalize_intensity);

cdef extern from "SpectralEntropy.h":
    float calculate_unweighted_entropy_similarity(
        float_spec *peaks_a, int peaks_a_len,
        float_spec *peaks_b, int peaks_b_len,
        float ms2_tolerance_in_da, float ms2_tolerance_in_ppm,
        bool clean_spectra,
        float min_mz, float max_mz,
        float noise_threshold,
        int max_peak_num);
    float calculate_entropy_similarity(
        float_spec *peaks_a, int peaks_a_len,
        float_spec *peaks_b, int peaks_b_len,
        float ms2_tolerance_in_da, float ms2_tolerance_in_ppm,
        bool clean_spectra,
        float min_mz, float max_mz,
        float noise_threshold,
        int max_peak_num);

cpdef np.ndarray[float32, ndim=2] cy_clean_spectrum(
    peaks,
    min_mz = -1,
    max_mz = -1,
    float noise_threshold = 0.01,
    float min_ms2_difference_in_da = 0.05,
    float min_ms2_difference_in_ppm = -1,
    max_peak_num = -1,
    int normalize_intensity = True,
):
    """
    Clean, centroid, and normalize a spectrum with the following steps:

        1. Remove empty peaks (m/z <= 0 or intensity <= 0).
        2. Remove peaks with m/z >= max_mz or m/z < min_mz.
        3. Centroid the spectrum by merging peaks within min_ms2_difference_in_da.
        4. Remove peaks with intensity < noise_threshold * max_intensity.
        5. Keep only the top max_peak_num peaks.
        6. Normalize the intensity to sum to 1.


    Parameters
    ----------
    peaks : np.ndarray in shape (n_peaks, 2), dtype=np.float32 or list[list[float, float]]
        A 2D array of shape (n_peaks, 2) where the first column is m/z and the second column is intensity.

    min_mz : float, optional
        The minimum m/z to keep. Defaults to None, which will skip removing peaks with m/z < min_mz.

    max_mz : float, optional
        The maximum m/z to keep. Defaults to None, which will skip removing peaks with m/z >= max_mz.

    noise_threshold : float, optional
        The minimum intensity to keep. Defaults to 0.01, which will remove peaks with intensity < 0.01 * max_intensity.

    min_ms2_difference_in_da : float, optional
        The minimum m/z difference between two peaks in the resulting spectrum. Defaults to 0.05, which will merge peaks within 0.05 Da. If a negative value is given, the min_ms2_difference_in_ppm will be used instead.

    min_ms2_difference_in_ppm : float, optional
        The minimum m/z difference between two peaks in the resulting spectrum. Defaults to -1, which will use the min_ms2_difference_in_da instead. If a negative value is given, the min_ms2_difference_in_da will be used instead.
        ** Note either min_ms2_difference_in_da or min_ms2_difference_in_ppm must be positive. If both are positive, min_ms2_difference_in_ppm will be used. **

    max_peak_num : int, optional
        The maximum number of peaks to keep. Defaults to None, which will keep all peaks.

    normalize_intensity : bool, optional
        Whether to normalize the intensity to sum to 1. Defaults to True. If False, the intensity will be kept as is.

        _

    Returns
    -------
    np.ndarray in shape (n_peaks, 2), dtype=np.float32
        The cleaned spectrum will be guaranteed to be sorted by m/z in ascending order.

    """
    if len(peaks) == 0:
        return np.zeros((0, 2), dtype=np.float32)
    if min_ms2_difference_in_da < 0 and min_ms2_difference_in_ppm < 0:
        raise ValueError("Either min_ms2_difference_in_da or min_ms2_difference_in_ppm must be positive.")

    cdef np.ndarray[float32, ndim=2] clean_peaks = np.array(peaks, dtype=np.float32, copy=True, order="C")
    min_mz = -1 if min_mz is None else min_mz
    max_mz = -1 if max_mz is None else max_mz
    max_peak_num = -1 if max_peak_num is None else max_peak_num

    cdef int clean_peaks_len = clean_spectrum(
        <float32*>clean_peaks.data,
        clean_peaks.shape[0],
        min_mz,
        max_mz,
        noise_threshold,
        min_ms2_difference_in_da,
        min_ms2_difference_in_ppm,
        max_peak_num,
        normalize_intensity,
    )
    if clean_peaks_len>0:
        clean_peaks = clean_peaks[:clean_peaks_len]
    else:
        clean_peaks = np.zeros((0, 2), dtype=np.float32)
    return clean_peaks


cpdef float cy_calculate_entropy_similarity(
    peaks_a, peaks_b,
    float ms2_tolerance_in_da = 0.02,
    float ms2_tolerance_in_ppm = -1,
    bool clean_spectra = True,
    float min_mz = -1,
    float max_mz = -1,
    float noise_threshold = 0.01,
    int max_peak_num = -1
):
    """Calculate the entropy similarity between two spectra.

    First, the spectra are cleaned by the `clean_spectrum()` function.

    Then, the entropy based intensity weights are applied to the peaks.

    Finally, the entropy similarity is calculated by the `calculate_unweighted_entropy_similarity()` function.

    The formula for entropy similarity is as follows:

    .. math::
        Similarity = \\frac{1}{2} \\begin{cases}
        0 & \\text{ if } mz_{A,i} \\neq mz_{B,j} \\\\
        \\sum_{i,j} {f(I_{A,i}+I_{B,j}) - f(I_{A,i}) - f(I_{B,j})} & \\text{ if } mz_{A,i} = mz_{B,j}
        \\end{cases}

    .. math::
        \\text{ where } f(x) = x \\log_2(x) \\text{ and } \\sum_{i} I_{A,i} = \\sum_{j} I_{B,j} = 1

    Parameters
    ----------
    peaks_a : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The first spectrum to calculate entropy similarity for. The first column is m/z, and the second column is intensity.

    peaks_b : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The second spectrum to calculate entropy similarity for. The first column is m/z, and the second column is intensity.

    ms2_tolerance_in_da : float, optional
        The MS2 tolerance in Da. Defaults to 0.02. If this is set to a negative value, ms2_tolerance_in_ppm will be used instead.

    ms2_tolerance_in_ppm : float, optional
        The MS2 tolerance in ppm. Defaults to -1. If this is set to a negative value, ms2_tolerance_in_da will be used instead.

        **Note:** Either `ms2_tolerance_in_da` or `ms2_tolerance_in_ppm` must be positive. If both `ms2_tolerance_in_da` and `ms2_tolerance_in_ppm` are positive, `ms2_tolerance_in_ppm` will be used.

    clean_spectra : bool, optional
        Whether to clean the spectra before calculating entropy similarity. Defaults to True. **Only set this to False if the spectra have been preprocessed by the `clean_spectrum()` function!** Otherwise, the results will be incorrect. If the spectra are already cleaned, set this to False to save time.

    **kwargs : optional
        The arguments and keyword arguments to pass to function ``clean_spectrum()``.

        _

    Returns
    -------
    float
        The entropy similarity between the two spectra.
    """
    if len(peaks_a) == 0 or len(peaks_b) == 0:
        return 0.0
    cdef np.ndarray[float32, ndim=2] peaks_a_np = np.array(peaks_a, dtype=np.float32, copy=True, order="C")
    cdef np.ndarray[float32, ndim=2] peaks_b_np = np.array(peaks_b, dtype=np.float32, copy=True, order="C")

    cdef float result = calculate_entropy_similarity(
        <float32*> peaks_a_np.data, len(peaks_a_np),
        <float32*> peaks_b_np.data, len(peaks_b_np),
        ms2_tolerance_in_da,
        ms2_tolerance_in_ppm,
        clean_spectra,
        min_mz,
        max_mz,
        noise_threshold,
        max_peak_num
    )
    return result


cpdef float cy_calculate_unweighted_entropy_similarity(
    peaks_a, peaks_b,
    float ms2_tolerance_in_da = 0.02,
    float ms2_tolerance_in_ppm = -1,
    bool clean_spectra = True,
    float min_mz = -1,
    float max_mz = -1,
    float noise_threshold = 0.01,
    int max_peak_num = -1
):
    """Calculate the unweighted entropy similarity between two spectra.

    The formula for unweighted entropy similarity is as follows:

    .. math::
        Similarity = \\frac{1}{2} \\begin{cases}
        0 & \\text{ if } mz_{A,i} \\neq mz_{B,j} \\\\
        \\sum_{i,j} {f(I_{A,i}+I_{B,j}) - f(I_{A,i}) - f(I_{B,j})} & \\text{ if } mz_{A,i} = mz_{B,j}
        \\end{cases}

    .. math::
        \\text{ where } f(x) = x \\log_2(x) \\text{ and } \\sum_{i} I_{A,i} = \\sum_{j} I_{B,j} = 1


    Parameters
    ----------
    peaks_a : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The first spectrum to calculate unweighted entropy similarity for. The first column is m/z, and the second column is intensity.

    peaks_b : np.ndarray in shape (n_peaks, 2), np.float32 or list[list[float, float]]
        The second spectrum to calculate unweighted entropy similarity for. The first column is m/z, and the second column is intensity.

    ms2_tolerance_in_da : float, optional
        The MS2 tolerance in Da. Defaults to 0.02. If this is set to a negative value, ms2_tolerance_in_ppm will be used instead.

    ms2_tolerance_in_ppm : float, optional
        The MS2 tolerance in ppm. Defaults to -1. If this is set to a negative value, ms2_tolerance_in_da will be used instead.

        **Note:** Either `ms2_tolerance_in_da` or `ms2_tolerance_in_ppm` must be positive. If both `ms2_tolerance_in_da` and `ms2_tolerance_in_ppm` are positive, `ms2_tolerance_in_ppm` will be used.

    clean_spectra : bool, optional
        Whether to clean the spectra before calculating unweighted entropy similarity. Defaults to True. Only set this to False if the spectra have been preprocessed by the clean_spectrum() function! Otherwise, the results will be incorrect. If the spectra are already cleaned, set this to False to save time. If the spectra are in the list format, always set this to True or an error will be raised.

    **kwargs : optional
        The arguments and keyword arguments to pass to function ``clean_spectrum()``.

        _

    Returns
    -------
    float
        The unweighted entropy similarity between the two spectra.
    """
    if len(peaks_a) == 0 or len(peaks_b) == 0:
        return 0.0
    cdef np.ndarray[float32, ndim=2] peaks_a_np
    cdef np.ndarray[float32, ndim=2] peaks_b_np
    if clean_spectra:
        peaks_a_np = np.array(peaks_a, dtype=np.float32, copy=True, order="C")
        peaks_b_np = np.array(peaks_b, dtype=np.float32, copy=True, order="C")
    else:
        peaks_a_np = np.array(peaks_a, dtype=np.float32, copy=False, order="C")
        peaks_b_np = np.array(peaks_b, dtype=np.float32, copy=False, order="C")
        
    cdef float result = calculate_unweighted_entropy_similarity(
        <float32*> peaks_a_np.data, len(peaks_a_np),
        <float32*> peaks_b_np.data, len(peaks_b_np),
        ms2_tolerance_in_da,
        ms2_tolerance_in_ppm,
        clean_spectra,
        min_mz,
        max_mz,
        noise_threshold,
        max_peak_num
    )
    return result



#!/usr/bin/env python3
import numpy as np


def clean_spectrum(
    peaks,
    min_mz: float = -1.0,
    max_mz: float = -1.0,
    noise_threshold: float = 0.01,
    min_ms2_difference_in_da: float = 0.05,
    min_ms2_difference_in_ppm: float = -1.0,
    max_peak_num: int = -1,
    normalize_intensity: bool = True,
    **kwargs,
) -> np.ndarray:
    """
    Clean, centroid, and normalize a spectrum with the following steps:

        1. Remove empty peaks (m/z <= 0 or intensity <= 0).
        2. Remove peaks with m/z >= max_mz or m/z < min_mz.
        3. Centroid the spectrum by merging peaks within min_ms2_difference_in_da.
        4. Remove peaks with intensity < noise_threshold * max_intensity.
        5. Keep only the top max_peak_num peaks.
        6. Normalize the intensity to sum to 1.


    Parameters
    ----------
    peaks : np.ndarray in shape (n_peaks, 2), dtype=np.float32 or list[list[float, float]]
        A 2D array of shape (n_peaks, 2) where the first column is m/z and the second column is intensity.

    min_mz : float, optional
        The minimum m/z to keep. Defaults to None, which will skip removing peaks with m/z < min_mz.

    max_mz : float, optional
        The maximum m/z to keep. Defaults to None, which will skip removing peaks with m/z >= max_mz.

    noise_threshold : float, optional
        The minimum intensity to keep. Defaults to 0.01, which will remove peaks with intensity < 0.01 * max_intensity.

    min_ms2_difference_in_da : float, optional
        The minimum m/z difference between two peaks in the resulting spectrum. Defaults to 0.05, which will merge peaks within 0.05 Da. If a negative value is given, the min_ms2_difference_in_ppm will be used instead.

    min_ms2_difference_in_ppm : float, optional
        The minimum m/z difference between two peaks in the resulting spectrum. Defaults to -1, which will use the min_ms2_difference_in_da instead. If a negative value is given, the min_ms2_difference_in_da will be used instead.
        ** Note either min_ms2_difference_in_da or min_ms2_difference_in_ppm must be positive. If both are positive, min_ms2_difference_in_ppm will be used. **

    max_peak_num : int, optional
        The maximum number of peaks to keep. Defaults to None, which will keep all peaks.

    normalize_intensity : bool, optional
        Whether to normalize the intensity to sum to 1. Defaults to True. If False, the intensity will be kept as is.

    **kwargs : optional
        Those keyword arguments will be ignored.

        _

    Returns
    -------
    np.ndarray in shape (n_peaks, 2), dtype=np.float32
        The cleaned spectrum will be guaranteed to be sorted by m/z in ascending order.

    """
    # Check the input spectrum and convert it to numpy array with shape (n, 2) and dtype np.float32.
    peaks = np.asarray(peaks, dtype=np.float32, order="C")
    if len(peaks) == 0:
        return np.zeros((0, 2), dtype=np.float32)
    assert peaks.ndim == 2 and peaks.shape[1] == 2, "The input spectrum must be a 2D numpy array with shape (n, 2)."

    # Step 1. Remove empty peaks (m/z <= 0 or intensity <= 0).
    peaks = peaks[np.bitwise_and(peaks[:, 0] > 0, peaks[:, 1] > 0)]

    # Step 2. Remove peaks with m/z >= max_mz or m/z < min_mz.
    if min_mz is not None and min_mz > 0:
        peaks = peaks[peaks[:, 0] >= min_mz]

    if max_mz is not None and max_mz > 0:
        peaks = peaks[peaks[:, 0] <= max_mz]

    if peaks.shape[0] == 0:
        return np.zeros((0, 2), dtype=np.float32)

    # Step 3. Centroid the spectrum by merging peaks within min_ms2_difference_in_da.
    # The peaks will be sorted by m/z in ascending order.
    if min_ms2_difference_in_ppm > 0:
        peaks = centroid_spectrum(peaks, ms2_da=-1, ms2_ppm=min_ms2_difference_in_ppm)
    elif min_ms2_difference_in_da > 0:
        peaks = centroid_spectrum(peaks, ms2_da=min_ms2_difference_in_da, ms2_ppm=-1)
    else:
        raise ValueError("Either min_ms2_difference_in_da or min_ms2_difference_in_ppm must be positive.")

    # Step 4. Remove peaks with intensity < noise_threshold * max_intensity.
    if noise_threshold is not None:
        peaks = peaks[peaks[:, 1] >= noise_threshold * np.max(peaks[:, 1])]

    # Step 5. Keep only the top max_peak_num peaks.
    if max_peak_num is not None and max_peak_num > 0:
        # Sort the spectrum by intensity.
        peaks = peaks[np.argsort(peaks[:, 1])[-max_peak_num:]]
        # Sort the spectrum by m/z.
        peaks = peaks[np.argsort(peaks[:, 0])]

    # Step 6. Normalize the intensity to sum to 1.
    if normalize_intensity:
        spectrum_sum = np.sum(peaks[:, 1])
        if spectrum_sum > 0:
            peaks[:, 1] /= spectrum_sum
        else:
            peaks = np.zeros((0, 2), dtype=np.float32)

    return peaks


def centroid_spectrum(peaks: np.ndarray, ms2_da: float = -1, ms2_ppm: float = -1) -> np.ndarray:
    """
    Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da.

    Parameters
    ----------
    peaks : np.ndarray in shape (n_peaks, 2), dtype=np.float32
        A 2D array of shape (n_peaks, 2) where the first column is m/z and the second column is intensity.

    ms2_ppm : float, optional
        The m/z tolerance in ppm. Defaults to -1, which will use ms2_da instead.

    ms2_da : float, optional
        The m/z tolerance in Da. Defaults to -1, which will use ms2_ppm instead.

        **Note: ms2_ppm and ms2_da cannot be both negative, if so, an Exception will be raised. If both are positive, ms2_da will be used.**

    Returns
    -------
    np.ndarray in shape (n_peaks, 2), dtype=np.float32
        The resulting spectrum after centroiding will be guarantee to have been sorted by m/z, and the difference between two peaks will be >= ms2_ppm or >= ms2_da.
    """

    # Sort the peaks by m/z.
    peaks = peaks[np.argsort(peaks[:, 0])]
    is_centroided: int = _check_centroid(peaks, ms2_da=ms2_da, ms2_ppm=ms2_ppm)
    while is_centroided == 0:
        peaks = _centroid_spectrum(peaks, ms2_da=ms2_da, ms2_ppm=ms2_ppm)
        is_centroided = _check_centroid(peaks, ms2_da=ms2_da, ms2_ppm=ms2_ppm)
    return peaks


def _centroid_spectrum(peaks: np.ndarray, ms2_da: float = -1, ms2_ppm: float = -1) -> np.ndarray:
    """Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da."""
    # Construct a new spectrum to avoid memory reallocation.
    peaks_new = np.zeros((peaks.shape[0], 2), dtype=np.float32)
    peaks_new_len = 0

    # Get the intensity argsort order.
    intensity_order = np.argsort(peaks[:, 1])

    mz_delta_allowed_left = ms2_da
    mz_delta_allowed_right = ms2_da
    # Iterate through the peaks from high to low intensity.
    for idx in intensity_order[::-1]:
        if ms2_ppm > 0:
            mz_delta_allowed_left = peaks[idx, 0] * ms2_ppm * 1e-6
            # For the right boundary, the mz_delta_allowed_right = peaks[right_idx, 0] * ms2_ppm * 1e-6 = peaks[idx, 0] / (1 - ms2_ppm * 1e-6) - peaks[idx, 0]
            mz_delta_allowed_right = peaks[idx, 0] / (1 - ms2_ppm * 1e-6) - peaks[idx, 0]

        if peaks[idx, 1] > 0:
            # Find the left boundary.
            left_idx = idx - 1
            while left_idx >= 0 and peaks[idx, 0] - peaks[left_idx, 0] <= mz_delta_allowed_left:
                left_idx -= 1
            left_idx += 1

            # Find the right boundary.
            right_idx = idx + 1
            while right_idx < peaks.shape[0] and peaks[right_idx, 0] - peaks[idx, 0] <= mz_delta_allowed_right:
                right_idx += 1

            # Merge the peaks within the boundary. The new m/z is the weighted average of the m/z and intensity.
            intensity_sum = np.sum(peaks[left_idx:right_idx, 1])
            intensity_weighted_mz_sum = np.sum(peaks[left_idx:right_idx, 1] * peaks[left_idx:right_idx, 0])
            peaks_new[peaks_new_len, 0] = intensity_weighted_mz_sum / intensity_sum
            peaks_new[peaks_new_len, 1] = intensity_sum
            peaks_new_len += 1

            # Set the intensity of the merged peaks to 0.
            peaks[left_idx:right_idx, 1] = 0

    # Return the new spectrum.
    peaks_new = peaks_new[:peaks_new_len]
    return peaks_new[np.argsort(peaks_new[:, 0])]


def _check_centroid(peaks: np.ndarray, ms2_da: float = -1, ms2_ppm: float = -1) -> int:
    """Check if the spectrum is centroided. 0 for False and 1 for True."""
    if peaks.shape[0] <= 1:
        return 1

    if ms2_ppm >= 0:
        # Use ms2_ppm to check if the spectrum is centroided.
        return 1 if np.all(np.diff(peaks[:, 0]) >= peaks[1:, 0] * ms2_ppm * 1e-6) else 0
    elif ms2_da >= 0:
        # Use ms2_da to check if the spectrum is centroided.
        return 1 if np.all(np.diff(peaks[:, 0]) >= ms2_da) else 0



from .flash_entropy_search import FlashEntropySearch
from .flash_entropy_search_core import FlashEntropySearchCore
from .flash_entropy_search_core_low_memory import FlashEntropySearchCoreLowMemory
from .flash_entropy_search_core_medium_memory import FlashEntropySearchCoreMediumMemory
from .flash_entropy_search_core_for_dynamic_indexing import FlashEntropySearchCoreForDynamicIndexing


#!/usr/bin/env python3
import numpy as np


def read_one_spectrum(file_input, **kwargs) -> dict:
    """
    Read all spectra from .lbm2 file.

    :param file_input: The file path of .lbm2 file.
    :return: A list of dict contains all spectra information.
    """
    try:
        import lz4.block
        import msgpack
    except ImportError:
        raise ImportError("Please install lz4 and msgpack package via `pip install lz4 msgpack`.")

    MSGPAK_INT32_LENGTH = 5
    with open(file_input, "rb") as fi:
        buffer = fi.read()
        buffer_data = msgpack.unpackb(buffer)
        lbm2_data = buffer_data.data

        lz4_data_length = msgpack.unpackb(lbm2_data[:MSGPAK_INT32_LENGTH])
        raw_buffer = lbm2_data[MSGPAK_INT32_LENGTH:]
        decompressed_buffer = lz4.block.decompress(raw_buffer, uncompressed_size=lz4_data_length)

        # print("Unpack all spectra")
        all_raw_spectra = msgpack.unpackb(decompressed_buffer)

    # Construct all spectra
    for i, spec_raw in enumerate(all_raw_spectra):
        adduct = spec_raw[13]  # AdductIonBean
        charge = {0: 1, 1: -1}[spec_raw[6]]  # IonMode
        charge = charge*adduct[3]  # ChargeNumber

        peaks = spec_raw[11]  # MzIntensityCommentBean
        peaks = (np.array(peaks)[:, 1:3]).astype(np.float32)

        spec = {
            "_ms_level": 2,
            "_scan_number": i+1, # ScanNumber starts from 1
            "id": spec_raw[0],  # Id
            "compound_class": spec_raw[2],  # CompoundClass
            "name": spec_raw[3],  # Name
            "rt": spec_raw[4],  # RetentionTime
            "smiles": spec_raw[7],  # Smiles
            "inchikey": spec_raw[8],  # InchiKey
            "precursor_mz": spec_raw[9],  # PrecursorMz
            "peaks": peaks,
            "charge": charge,
            "adduct": adduct[2],
        }
        yield spec
    return all_raw_spectra


if __name__ == "__main__":
    file_input = "/p/FastEntropySearch/next_version/io/test.lbm2"
    all_spectra = [x for x in read_one_spectrum(file_input)]



import numpy as np

try:
    from .fast_flash_entropy_search_cpython import  cy_entropy_similarity_identity_search as entropy_similarity_search_identity

except ImportError:
    def entropy_similarity_search_identity(
        product_mz_idx_min,
        product_mz_idx_max,
        intensity,
        entropy_similarity,
        library_peaks_intensity,
        library_spec_idx_array,
        search_spectra_idx_min,
        search_spectra_idx_max,
    ):
        """
        The entropy_similarity will be modified in this function.
        search_type is 0: search all spectra.
        search_type is 1: search spectra in the range [search_spectra_idx_min, search_spectra_idx_max).

        Note: the intensity here should be half of the original intensity.
        """
        all_library_spec_idx = library_spec_idx_array[product_mz_idx_min:product_mz_idx_max]
        idx_list = product_mz_idx_min + np.where(np.bitwise_and(all_library_spec_idx >= search_spectra_idx_min,
                                                 all_library_spec_idx < search_spectra_idx_max))[0]

        array_library_spec_idx = library_spec_idx_array[idx_list]
        array_library_peak_intensity = library_peaks_intensity[idx_list]

        array_library_ab = intensity + array_library_peak_intensity
        entropy_similarity[array_library_spec_idx] += (
            array_library_ab * np.log2(array_library_ab) - intensity * np.log2(intensity) - array_library_peak_intensity * np.log2(array_library_peak_intensity)
        )



#!/usr/bin/env python3
import json
import numpy as np
from pathlib import Path
from .flash_entropy_search_core import FlashEntropySearchCore


class FlashEntropySearchCoreMediumMemory(FlashEntropySearchCore):
    def __init__(self, path_data, max_ms2_tolerance_in_da=0.024, mz_index_step=0.0001, intensity_weight="entropy") -> None:
        """
        Initialize the EntropySearch class.
        This class is use memmap function to read data from the disk, which is suitable for most of the cases, unless the data is super large.
        
        :param path_data:   The path to save the index data.
        :param max_ms2_tolerance_in_da:  The maximum MS2 tolerance in Da.
        :param mz_index_step:   The step size for the m/z index.
        :param intensity_weight:    The weight for the intensity in the entropy calculation, can be "entropy" or None. Default is "entropy".
            - None: The intensity will not be weighted, then the unweighted similarity will be calculated.
            - "entropy": The intensity will be weighted by the entropy, then the entropy similarity will be calculated.
        """
        super().__init__(max_ms2_tolerance_in_da=max_ms2_tolerance_in_da, mz_index_step=mz_index_step, intensity_weight=intensity_weight)
        self.path_data = Path(str(path_data))
        self.path_data.mkdir(parents=True, exist_ok=True)

    def _generate_index_from_peak_data(self, peak_data, max_indexed_mz, append):
        total_peaks_num = peak_data.shape[0]

        # Sort with precursor m/z.
        peak_data.sort(order="ion_mz")

        # Record the m/z, intensity, and spectrum index information for product ions.
        (peak_data["ion_mz"]).tofile(self.path_data / "all_ions_mz.npy")
        (peak_data["intensity"]).tofile(self.path_data / "all_ions_intensity.npy")
        (peak_data["spec_idx"]).tofile(self.path_data / "all_ions_spec_idx.npy")

        # all_ions_mz = self._convert_view_to_array(peak_data.view(np.float32).reshape(total_peaks_num, -1)[:, 0], np.float32, "all_ions_mz")
        # all_ions_intensity = self._convert_view_to_array(peak_data.view(np.float32).reshape(total_peaks_num, -1)[:, 2], np.float32, "all_ions_intensity")
        # all_ions_spec_idx = self._convert_view_to_array(peak_data.view(np.uint32).reshape(total_peaks_num, -1)[:, 3], np.uint32, "all_ions_spec_idx")

        # Assign the index of the product ions.
        peak_data["peak_idx"] = np.arange(0, self.total_peaks_num, dtype=np.uint64)

        # Build index for fast access to the ion's m/z.
        all_ions_mz = np.memmap(self.path_data / "all_ions_mz.npy", dtype=np.float32, mode="r", shape=(total_peaks_num,))
        max_mz = min(np.max(all_ions_mz), max_indexed_mz)
        search_array = np.arange(0.0, max_mz, self.mz_index_step)
        all_ions_mz_idx_start = np.searchsorted(all_ions_mz, search_array, side="left").astype(np.int64)
        all_ions_mz_idx_start.tofile(self.path_data / "all_ions_mz_idx_start.npy")

        ############## Step 3: Build the index by sort with neutral loss mass. ##############
        # Sort with the neutral loss mass.
        peak_data.sort(order="nl_mass")

        # Record the m/z, intensity, spectrum index, and product ions index information for neutral loss ions.
        (peak_data["nl_mass"]).tofile(self.path_data / "all_nl_mass.npy")
        (peak_data["intensity"]).tofile(self.path_data / "all_nl_intensity.npy")
        (peak_data["spec_idx"]).tofile(self.path_data / "all_nl_spec_idx.npy")
        (peak_data["peak_idx"]).tofile(self.path_data / "all_ions_idx_for_nl.npy")

        # all_nl_mass = self._convert_view_to_array(peak_data.view(np.float32).reshape(total_peaks_num, -1)[:, 1], np.float32, "all_nl_mass")
        # all_nl_intensity = self._convert_view_to_array(peak_data.view(np.float32).reshape(total_peaks_num, -1)[:, 2], np.float32, "all_nl_intensity")
        # all_nl_spec_idx = self._convert_view_to_array(peak_data.view(np.uint32).reshape(total_peaks_num, -1)[:, 3], np.uint32, "all_nl_spec_idx")
        # all_ions_idx_for_nl = self._convert_view_to_array(peak_data.view(np.uint64).reshape(total_peaks_num, -1)[:, 2], np.uint64, "all_ions_idx_for_nl")

        # Build the index for fast access to the neutral loss mass.
        all_nl_mass = np.memmap(self.path_data / "all_nl_mass.npy", dtype=np.float32, mode="r", shape=(total_peaks_num,))
        max_mz = min(np.max(all_nl_mass), max_indexed_mz)
        search_array = np.arange(0.0, max_mz, self.mz_index_step)
        all_nl_mass_idx_start = np.searchsorted(all_nl_mass, search_array, side="left").astype(np.int64)
        all_nl_mass_idx_start.tofile(self.path_data / "all_nl_mass_idx_start.npy")

        ############## Step 4: Save the index. ##############
        self.write()
        self.read()
        return self.index

    def read(self, path_data=None):
        """
        Read the index from the file.
        """
        if path_data is not None:
            self.path_data = Path(path_data)

        try:
            self.index = []
            for name in self.index_names:
                self.index.append(np.memmap(self.path_data / f"{name}.npy", dtype=self.index_dtypes[name], mode="r"))

            with open(self.path_data / "information.json", "r") as f:
                information = json.load(f)
            self.mz_index_step = information["mz_index_step"]
            self.total_spectra_num = information["total_spectra_num"]
            self.total_peaks_num = information["total_peaks_num"]
            self.max_ms2_tolerance_in_da = information["max_ms2_tolerance_in_da"]
            return True
        except:
            return False

    def write(self, path_data=None):
        """
        Write the index to the file.
        """
        if path_data is not None:
            assert Path(path_data) == self.path_data, "The path_data is not the same as the path_data in the class."

        information = {
            "mz_index_step": float(self.mz_index_step),
            "total_spectra_num": int(self.total_spectra_num),
            "total_peaks_num": int(self.total_peaks_num),
            "max_ms2_tolerance_in_da": float(self.max_ms2_tolerance_in_da),
        }
        json.dump(information, open(self.path_data / "information.json", "w"))
